"""Test transaction."""
from binascii import a2b_hex
from itertools import count
import unittest
from unittest.mock import MagicMock, patch

from pymodbus.exceptions import (
    InvalidMessageReceivedException,
    ModbusIOException,
)
from pymodbus.factory import ServerDecoder
from pymodbus.pdu import ModbusRequest
from pymodbus.transaction import (
    DictTransactionManager,
    FifoTransactionManager,
    ModbusAsciiFramer,
    ModbusBinaryFramer,
    ModbusRtuFramer,
    ModbusSocketFramer,
    ModbusTlsFramer,
    ModbusTransactionManager,
)


TEST_MESSAGE = b"\x7b\x01\x03\x00\x00\x00\x05\x85\xC9\x7d"


class ModbusTransactionTest(  # pylint: disable=too-many-public-methods
    unittest.TestCase
):
    """Unittest for the pymodbus.transaction module."""

    # ----------------------------------------------------------------------- #
    # Test Construction
    # ----------------------------------------------------------------------- #
    def setUp(self):
        """Set up the test environment"""
        self.client = None
        self.decoder = ServerDecoder()
        self._tcp = ModbusSocketFramer(decoder=self.decoder, client=None)
        self._tls = ModbusTlsFramer(decoder=self.decoder, client=None)
        self._rtu = ModbusRtuFramer(decoder=self.decoder, client=None)
        self._ascii = ModbusAsciiFramer(decoder=self.decoder, client=None)
        self._binary = ModbusBinaryFramer(decoder=self.decoder, client=None)
        self._manager = DictTransactionManager(self.client)
        self._queue_manager = FifoTransactionManager(self.client)
        self._tm = ModbusTransactionManager(self.client)

    def tearDown(self):
        """Clean up the test environment"""
        del self._manager
        del self._tcp
        del self._tls
        del self._rtu
        del self._ascii

    # ----------------------------------------------------------------------- #
    # Base transaction manager
    # ----------------------------------------------------------------------- #

    def test_calculate_expected_response_length(self):
        """Test calculate expected response length."""
        self._tm.client = MagicMock()
        self._tm.client.framer = MagicMock()
        self._tm._set_adu_size()  # pylint: disable=protected-access
        self.assertEqual(
            self._tm._calculate_response_length(0),  # pylint: disable=protected-access
            None,
        )
        self._tm.base_adu_size = 10
        self.assertEqual(
            self._tm._calculate_response_length(5),  # pylint: disable=protected-access
            15,
        )

    def test_calculate_exception_length(self):
        """Test calculate exception length."""
        for framer, exception_length in (
            ("ascii", 11),
            ("binary", 7),
            ("rtu", 5),
            ("tcp", 9),
            ("tls", 2),
            ("dummy", None),
        ):
            self._tm.client = MagicMock()
            if framer == "ascii":
                self._tm.client.framer = self._ascii
            elif framer == "binary":
                self._tm.client.framer = self._binary
            elif framer == "rtu":
                self._tm.client.framer = self._rtu
            elif framer == "tcp":
                self._tm.client.framer = self._tcp
            elif framer == "tls":
                self._tm.client.framer = self._tls
            else:
                self._tm.client.framer = MagicMock()

            self._tm._set_adu_size()  # pylint: disable=protected-access
            self.assertEqual(
                self._tm._calculate_exception_length(),  # pylint: disable=protected-access
                exception_length,
            )

    @patch("pymodbus.transaction.time")
    def test_execute(self, mock_time):
        """Test execute."""
        mock_time.time.side_effect = count()

        client = MagicMock()
        client.framer = self._ascii
        client.framer._buffer = b"deadbeef"  # pylint: disable=protected-access
        client.framer.processIncomingPacket = MagicMock()
        client.framer.processIncomingPacket.return_value = None
        client.framer.buildPacket = MagicMock()
        client.framer.buildPacket.return_value = b"deadbeef"
        client.framer.sendPacket = MagicMock()
        client.framer.sendPacket.return_value = len(b"deadbeef")
        client.framer.decode_data = MagicMock()
        client.framer.decode_data.return_value = {"unit": 1, "fcode": 222, "length": 27}
        request = MagicMock()
        request.get_response_pdu_size.return_value = 10
        request.unit_id = 1
        request.function_code = 222
        trans = ModbusTransactionManager(client)
        trans._recv = MagicMock(  # pylint: disable=protected-access
            return_value=b"abcdef"
        )
        self.assertEqual(trans.retries, 3)
        self.assertEqual(trans.retry_on_empty, False)

        trans.getTransaction = MagicMock()
        trans.getTransaction.return_value = "response"
        response = trans.execute(request)
        self.assertEqual(response, "response")
        # No response
        trans._recv = MagicMock(  # pylint: disable=protected-access
            return_value=b"abcdef"
        )
        trans.transactions = []
        trans.getTransaction = MagicMock()
        trans.getTransaction.return_value = None
        response = trans.execute(request)
        self.assertIsInstance(response, ModbusIOException)

        # No response with retries
        trans.retry_on_empty = True
        trans._recv = MagicMock(  # pylint: disable=protected-access
            side_effect=iter([b"", b"abcdef"])
        )
        response = trans.execute(request)
        self.assertIsInstance(response, ModbusIOException)

        # wrong handle_local_echo
        trans._recv = MagicMock(  # pylint: disable=protected-access
            side_effect=iter([b"abcdef", b"deadbe", b"123456"])
        )
        client.handle_local_echo = True
        trans.retry_on_empty = False
        trans.retry_on_invalid = False
        self.assertEqual(
            trans.execute(request).message, "[Input/Output] Wrong local echo"
        )
        client.handle_local_echo = False

        # retry on invalid response
        trans.retry_on_invalid = True
        trans._recv = MagicMock(  # pylint: disable=protected-access
            side_effect=iter([b"", b"abcdef", b"deadbe", b"123456"])
        )
        response = trans.execute(request)
        self.assertIsInstance(response, ModbusIOException)

        # Unable to decode response
        trans._recv = MagicMock(  # pylint: disable=protected-access
            side_effect=ModbusIOException()
        )
        client.framer.processIncomingPacket.side_effect = MagicMock(
            side_effect=ModbusIOException()
        )
        self.assertIsInstance(trans.execute(request), ModbusIOException)

        # Broadcast
        client.params.broadcast_enable = True
        request.unit_id = 0
        response = trans.execute(request)
        self.assertEqual(response, b"Broadcast write sent - no response expected")

    # ----------------------------------------------------------------------- #
    # Dictionary based transaction manager
    # ----------------------------------------------------------------------- #

    def test_dict_transaction_manager_tid(self):
        """Test the dict transaction manager TID"""
        for tid in range(1, self._manager.getNextTID() + 10):
            self.assertEqual(tid + 1, self._manager.getNextTID())
        self._manager.reset()
        self.assertEqual(1, self._manager.getNextTID())

    def test_get_dict_fifo_transaction_manager_transaction(self):
        """Test the dict transaction manager"""

        class Request:  # pylint: disable=too-few-public-methods
            """Request."""

        self._manager.reset()
        handle = Request()
        handle.transaction_id = (  # pylint: disable=attribute-defined-outside-init
            self._manager.getNextTID()
        )
        handle.message = b"testing"  # pylint: disable=attribute-defined-outside-init
        self._manager.addTransaction(handle)
        result = self._manager.getTransaction(handle.transaction_id)
        self.assertEqual(handle.message, result.message)

    def test_delete_dict_fifo_transaction_manager_transaction(self):
        """Test the dict transaction manager"""

        class Request:  # pylint: disable=too-few-public-methods
            """Request."""

        self._manager.reset()
        handle = Request()
        handle.transaction_id = (  # pylint: disable=attribute-defined-outside-init
            self._manager.getNextTID()
        )
        handle.message = b"testing"  # pylint: disable=attribute-defined-outside-init

        self._manager.addTransaction(handle)
        self._manager.delTransaction(handle.transaction_id)
        self.assertEqual(None, self._manager.getTransaction(handle.transaction_id))

    # ----------------------------------------------------------------------- #
    # Queue based transaction manager
    # ----------------------------------------------------------------------- #
    def test_fifo_transaction_manager_tid(self):
        """Test the fifo transaction manager TID"""
        for tid in range(1, self._queue_manager.getNextTID() + 10):
            self.assertEqual(tid + 1, self._queue_manager.getNextTID())
        self._queue_manager.reset()
        self.assertEqual(1, self._queue_manager.getNextTID())

    def test_get_fifo_transaction_manager_transaction(self):
        """Test the fifo transaction manager"""

        class Request:  # pylint: disable=too-few-public-methods
            """Request."""

        self._queue_manager.reset()
        handle = Request()
        handle.transaction_id = (  # pylint: disable=attribute-defined-outside-init
            self._queue_manager.getNextTID()
        )
        handle.message = b"testing"  # pylint: disable=attribute-defined-outside-init
        self._queue_manager.addTransaction(handle)
        result = self._queue_manager.getTransaction(handle.transaction_id)
        self.assertEqual(handle.message, result.message)

    def test_delete_fifo_transaction_manager_transaction(self):
        """Test the fifo transaction manager"""

        class Request:  # pylint: disable=too-few-public-methods
            """Request."""

        self._queue_manager.reset()
        handle = Request()
        handle.transaction_id = (  # pylint: disable=attribute-defined-outside-init
            self._queue_manager.getNextTID()
        )
        handle.message = b"testing"  # pylint: disable=attribute-defined-outside-init

        self._queue_manager.addTransaction(handle)
        self._queue_manager.delTransaction(handle.transaction_id)
        self.assertEqual(
            None, self._queue_manager.getTransaction(handle.transaction_id)
        )

    # ----------------------------------------------------------------------- #
    # TCP tests
    # ----------------------------------------------------------------------- #
    def test_tcp_framer_transaction_ready(self):
        """Test a tcp frame transaction"""
        msg = b"\x00\x01\x12\x34\x00\x04\xff\x02\x12\x34"
        self.assertFalse(self._tcp.isFrameReady())
        self.assertFalse(self._tcp.checkFrame())
        self._tcp.addToFrame(msg)
        self.assertTrue(self._tcp.isFrameReady())
        self.assertTrue(self._tcp.checkFrame())
        self._tcp.advanceFrame()
        self.assertFalse(self._tcp.isFrameReady())
        self.assertFalse(self._tcp.checkFrame())
        self.assertEqual(b"", self._ascii.getFrame())

    def test_tcp_framer_transaction_full(self):
        """Test a full tcp frame transaction"""
        msg = b"\x00\x01\x12\x34\x00\x04\xff\x02\x12\x34"
        self._tcp.addToFrame(msg)
        self.assertTrue(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg[7:], result)
        self._tcp.advanceFrame()

    def test_tcp_framer_transaction_half(self):
        """Test a half completed tcp frame transaction"""
        msg1 = b"\x00\x01\x12\x34\x00"
        msg2 = b"\x04\xff\x02\x12\x34"
        self._tcp.addToFrame(msg1)
        self.assertFalse(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(b"", result)
        self._tcp.addToFrame(msg2)
        self.assertTrue(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg2[2:], result)
        self._tcp.advanceFrame()

    def test_tcp_framer_transaction_half2(self):
        """Test a half completed tcp frame transaction"""
        msg1 = b"\x00\x01\x12\x34\x00\x04\xff"
        msg2 = b"\x02\x12\x34"
        self._tcp.addToFrame(msg1)
        self.assertFalse(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(b"", result)
        self._tcp.addToFrame(msg2)
        self.assertTrue(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg2, result)
        self._tcp.advanceFrame()

    def test_tcp_framer_transaction_half3(self):
        """Test a half completed tcp frame transaction"""
        msg1 = b"\x00\x01\x12\x34\x00\x04\xff\x02\x12"
        msg2 = b"\x34"
        self._tcp.addToFrame(msg1)
        self.assertFalse(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg1[7:], result)
        self._tcp.addToFrame(msg2)
        self.assertTrue(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg1[7:] + msg2, result)
        self._tcp.advanceFrame()

    def test_tcp_framer_transaction_short(self):
        """Test that we can get back on track after an invalid message"""
        msg1 = b"\x99\x99\x99\x99\x00\x01\x00\x01"
        msg2 = b"\x00\x01\x12\x34\x00\x04\xff\x02\x12\x34"
        self._tcp.addToFrame(msg1)
        self.assertFalse(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(b"", result)
        self._tcp.advanceFrame()
        self._tcp.addToFrame(msg2)
        self.assertEqual(10, len(self._tcp._buffer))  # pylint: disable=protected-access
        self.assertTrue(self._tcp.checkFrame())
        result = self._tcp.getFrame()
        self.assertEqual(msg2[7:], result)
        self._tcp.advanceFrame()

    def test_tcp_framer_populate(self):
        """Test a tcp frame packet build"""
        expected = ModbusRequest()
        expected.transaction_id = 0x0001
        expected.protocol_id = 0x1234
        expected.unit_id = 0xFF
        msg = b"\x00\x01\x12\x34\x00\x04\xff\x02\x12\x34"
        self._tcp.addToFrame(msg)
        self.assertTrue(self._tcp.checkFrame())
        actual = ModbusRequest()
        self._tcp.populateResult(actual)
        for name in ("transaction_id", "protocol_id", "unit_id"):
            self.assertEqual(getattr(expected, name), getattr(actual, name))
        self._tcp.advanceFrame()

    def test_tcp_framer_packet(self):
        """Test a tcp frame packet build"""
        old_encode = ModbusRequest.encode
        ModbusRequest.encode = lambda self: b""
        message = ModbusRequest()
        message.transaction_id = 0x0001
        message.protocol_id = 0x1234
        message.unit_id = 0xFF
        message.function_code = 0x01
        expected = b"\x00\x01\x12\x34\x00\x02\xff\x01"
        actual = self._tcp.buildPacket(message)
        self.assertEqual(expected, actual)
        ModbusRequest.encode = old_encode

    # ----------------------------------------------------------------------- #
    # TLS tests
    # ----------------------------------------------------------------------- #
    def test_framer_tls_framer_transaction_ready(self):
        """Test a tls frame transaction"""
        msg = b"\x01\x12\x34\x00\x08"
        self.assertFalse(self._tls.isFrameReady())
        self.assertFalse(self._tls.checkFrame())
        self._tls.addToFrame(msg)
        self.assertTrue(self._tls.isFrameReady())
        self.assertTrue(self._tls.checkFrame())
        self._tls.advanceFrame()
        self.assertFalse(self._tls.isFrameReady())
        self.assertFalse(self._tls.checkFrame())
        self.assertEqual(b"", self._tls.getFrame())

    def test_framer_tls_framer_transaction_full(self):
        """Test a full tls frame transaction"""
        msg = b"\x01\x12\x34\x00\x08"
        self._tls.addToFrame(msg)
        self.assertTrue(self._tls.checkFrame())
        result = self._tls.getFrame()
        self.assertEqual(msg[0:], result)
        self._tls.advanceFrame()

    def test_framer_tls_framer_transaction_half(self):
        """Test a half completed tls frame transaction"""
        msg1 = b""
        msg2 = b"\x01\x12\x34\x00\x08"
        self._tls.addToFrame(msg1)
        self.assertFalse(self._tls.checkFrame())
        result = self._tls.getFrame()
        self.assertEqual(b"", result)
        self._tls.addToFrame(msg2)
        self.assertTrue(self._tls.checkFrame())
        result = self._tls.getFrame()
        self.assertEqual(msg2[0:], result)
        self._tls.advanceFrame()

    def test_framer_tls_framer_transaction_short(self):
        """Test that we can get back on track after an invalid message"""
        msg1 = b""
        msg2 = b"\x01\x12\x34\x00\x08"
        self._tls.addToFrame(msg1)
        self.assertFalse(self._tls.checkFrame())
        result = self._tls.getFrame()
        self.assertEqual(b"", result)
        self._tls.advanceFrame()
        self._tls.addToFrame(msg2)
        self.assertEqual(5, len(self._tls._buffer))  # pylint: disable=protected-access
        self.assertTrue(self._tls.checkFrame())
        result = self._tls.getFrame()
        self.assertEqual(msg2[0:], result)
        self._tls.advanceFrame()

    def test_framer_tls_framer_decode(self):
        """Testmessage decoding"""
        msg1 = b""
        msg2 = b"\x01\x12\x34\x00\x08"
        result = self._tls.decode_data(msg1)
        self.assertEqual({}, result)
        result = self._tls.decode_data(msg2)
        self.assertEqual({"fcode": 1}, result)
        self._tls.advanceFrame()

    def test_framer_tls_incoming_packet(self):
        """Framer tls incoming packet."""
        msg = b"\x01\x12\x34\x00\x08"

        unit = 0x01

        def mock_callback(self):  # pylint: disable=unused-argument
            """Mock callback."""

        self._tls._process = MagicMock()  # pylint: disable=protected-access
        self._tls.isFrameReady = MagicMock(return_value=False)
        self._tls.processIncomingPacket(msg, mock_callback, unit)
        self.assertEqual(msg, self._tls.getRawFrame())
        self._tls.advanceFrame()

        self._tls.isFrameReady = MagicMock(return_value=True)
        self._tls._validate_unit_id = MagicMock(  # pylint: disable=protected-access
            return_value=False
        )
        self._tls.processIncomingPacket(msg, mock_callback, unit)
        self.assertEqual(b"", self._tls.getRawFrame())
        self._tls.advanceFrame()

        self._tls._validate_unit_id = MagicMock(  # pylint: disable=protected-access
            return_value=True
        )
        self._tls.processIncomingPacket(msg, mock_callback, unit)
        self.assertEqual(msg, self._tls.getRawFrame())
        self._tls.advanceFrame()

    def test_framer_tls_process(self):
        """Framer tls process."""

        class MockResult:  # pylint: disable=too-few-public-methods
            """Mock result."""

            def __init__(self, code):
                """Init."""
                self.function_code = code

        def mock_callback(self):  # pylint: disable=unused-argument
            """Mock callback."""

        self._tls.decoder.decode = MagicMock(return_value=None)
        self.assertRaises(
            ModbusIOException,
            lambda: self._tls._process(  # pylint: disable=protected-access
                mock_callback
            ),
        )

        result = MockResult(0x01)
        self._tls.decoder.decode = MagicMock(return_value=result)
        self.assertRaises(
            InvalidMessageReceivedException,
            lambda: self._tls._process(  # pylint: disable=protected-access
                mock_callback, error=True
            ),
        )

        self._tls._process(mock_callback)  # pylint: disable=protected-access
        self.assertEqual(b"", self._tls.getRawFrame())

    def test_framer_tls_framer_populate(self):
        """Test a tls frame packet build"""
        ModbusRequest()
        msg = b"\x01\x12\x34\x00\x08"
        self._tls.addToFrame(msg)
        self.assertTrue(self._tls.checkFrame())
        actual = ModbusRequest()
        result = self._tls.populateResult(  # pylint: disable=assignment-from-none
            actual
        )
        self.assertEqual(None, result)
        self._tls.advanceFrame()

    def test_framer_tls_framer_packet(self):
        """Test a tls frame packet build"""
        old_encode = ModbusRequest.encode
        ModbusRequest.encode = lambda self: b""
        message = ModbusRequest()
        message.function_code = 0x01
        expected = b"\x01"
        actual = self._tls.buildPacket(message)
        self.assertEqual(expected, actual)
        ModbusRequest.encode = old_encode

    # ----------------------------------------------------------------------- #
    # RTU tests
    # ----------------------------------------------------------------------- #
    def test_rtu_framer_transaction_ready(self):
        """Test if the checks for a complete frame work"""
        self.assertFalse(self._rtu.isFrameReady())

        msg_parts = [b"\x00\x01\x00", b"\x00\x00\x01\xfc\x1b"]
        self._rtu.addToFrame(msg_parts[0])
        self.assertFalse(self._rtu.isFrameReady())
        self.assertFalse(self._rtu.checkFrame())

        self._rtu.addToFrame(msg_parts[1])
        self.assertTrue(self._rtu.isFrameReady())
        self.assertTrue(self._rtu.checkFrame())

    def test_rtu_framer_transaction_full(self):
        """Test a full rtu frame transaction"""
        msg = b"\x00\x01\x00\x00\x00\x01\xfc\x1b"
        stripped_msg = msg[1:-2]
        self._rtu.addToFrame(msg)
        self.assertTrue(self._rtu.checkFrame())
        result = self._rtu.getFrame()
        self.assertEqual(stripped_msg, result)
        self._rtu.advanceFrame()

    def test_rtu_framer_transaction_half(self):
        """Test a half completed rtu frame transaction"""
        msg_parts = [b"\x00\x01\x00", b"\x00\x00\x01\xfc\x1b"]
        stripped_msg = b"".join(msg_parts)[1:-2]
        self._rtu.addToFrame(msg_parts[0])
        self.assertFalse(self._rtu.checkFrame())
        self._rtu.addToFrame(msg_parts[1])
        self.assertTrue(self._rtu.isFrameReady())
        self.assertTrue(self._rtu.checkFrame())
        result = self._rtu.getFrame()
        self.assertEqual(stripped_msg, result)
        self._rtu.advanceFrame()

    def test_rtu_framer_populate(self):
        """Test a rtu frame packet build"""
        request = ModbusRequest()
        msg = b"\x00\x01\x00\x00\x00\x01\xfc\x1b"
        self._rtu.addToFrame(msg)
        self._rtu.populateHeader()
        self._rtu.populateResult(request)

        header_dict = self._rtu._header  # pylint: disable=protected-access
        self.assertEqual(len(msg), header_dict["len"])
        self.assertEqual(int(msg[0]), header_dict["uid"])
        self.assertEqual(msg[-2:], header_dict["crc"])

        self.assertEqual(0x00, request.unit_id)

    def test_rtu_framer_packet(self):
        """Test a rtu frame packet build"""
        old_encode = ModbusRequest.encode
        ModbusRequest.encode = lambda self: b""
        message = ModbusRequest()
        message.unit_id = 0xFF
        message.function_code = 0x01
        expected = b"\xff\x01\x81\x80"  # only header + CRC - no data
        actual = self._rtu.buildPacket(message)
        self.assertEqual(expected, actual)
        ModbusRequest.encode = old_encode

    def test_rtu_decode_exception(self):
        """Test that the RTU framer can decode errors"""
        message = b"\x00\x90\x02\x9c\x01"
        self._rtu.addToFrame(message)
        result = self._rtu.checkFrame()
        self.assertTrue(result)

    def test_process(self):
        """Test process."""

        class MockResult:  # pylint: disable=too-few-public-methods
            """Mock result."""

            def __init__(self, code):
                self.function_code = code

        def mock_callback(self):  # pylint: disable=unused-argument
            """Mock callback."""

        mock_result = MockResult(code=0)
        self._rtu.getRawFrame = self._rtu.getFrame = MagicMock()
        self._rtu.decoder = MagicMock()
        self._rtu.decoder.decode = MagicMock(return_value=mock_result)
        self._rtu.populateResult = MagicMock()
        self._rtu.advanceFrame = MagicMock()

        self._rtu._process(mock_callback)  # pylint: disable=protected-access
        self._rtu.populateResult.assert_called_with(mock_result)
        self._rtu.advanceFrame.assert_called_with()
        self.assertTrue(self._rtu.advanceFrame.called)

        # Check errors
        self._rtu.decoder.decode = MagicMock(return_value=None)
        self.assertRaises(
            ModbusIOException,
            lambda: self._rtu._process(  # pylint: disable=protected-access
                mock_callback
            ),
        )

    def test_rtu_process_incoming_packets(self):
        """Test rtu process incoming packets."""
        mock_data = b"\x00\x01\x00\x00\x00\x01\xfc\x1b"
        unit = 0x00

        def mock_callback(self):  # pylint: disable=unused-argument
            """Mock callback."""

        self._rtu.addToFrame = MagicMock()
        self._rtu._process = MagicMock()  # pylint: disable=protected-access
        self._rtu.isFrameReady = MagicMock(return_value=False)
        self._rtu._buffer = mock_data  # pylint: disable=protected-access

        self._rtu.processIncomingPacket(mock_data, mock_callback, unit)

    # ----------------------------------------------------------------------- #
    # ASCII tests
    # ----------------------------------------------------------------------- #
    def test_ascii_framer_transaction_ready(self):
        """Test a ascii frame transaction"""
        msg = b":F7031389000A60\r\n"
        self.assertFalse(self._ascii.isFrameReady())
        self.assertFalse(self._ascii.checkFrame())
        self._ascii.addToFrame(msg)
        self.assertTrue(self._ascii.isFrameReady())
        self.assertTrue(self._ascii.checkFrame())
        self._ascii.advanceFrame()
        self.assertFalse(self._ascii.isFrameReady())
        self.assertFalse(self._ascii.checkFrame())
        self.assertEqual(b"", self._ascii.getFrame())

    def test_ascii_framer_transaction_full(self):
        """Test a full ascii frame transaction"""
        msg = b"sss:F7031389000A60\r\n"
        pack = a2b_hex(msg[6:-4])
        self._ascii.addToFrame(msg)
        self.assertTrue(self._ascii.checkFrame())
        result = self._ascii.getFrame()
        self.assertEqual(pack, result)
        self._ascii.advanceFrame()

    def test_ascii_framer_transaction_half(self):
        """Test a half completed ascii frame transaction"""
        msg1 = b"sss:F7031389"
        msg2 = b"000A60\r\n"
        pack = a2b_hex(msg1[6:] + msg2[:-4])
        self._ascii.addToFrame(msg1)
        self.assertFalse(self._ascii.checkFrame())
        result = self._ascii.getFrame()
        self.assertEqual(b"", result)
        self._ascii.addToFrame(msg2)
        self.assertTrue(self._ascii.checkFrame())
        result = self._ascii.getFrame()
        self.assertEqual(pack, result)
        self._ascii.advanceFrame()

    def test_ascii_framer_populate(self):
        """Test a ascii frame packet build"""
        request = ModbusRequest()
        self._ascii.populateResult(request)
        self.assertEqual(0x00, request.unit_id)

    def test_ascii_framer_packet(self):
        """Test a ascii frame packet build"""
        old_encode = ModbusRequest.encode
        ModbusRequest.encode = lambda self: b""
        message = ModbusRequest()
        message.unit_id = 0xFF
        message.function_code = 0x01
        expected = b":FF0100\r\n"
        actual = self._ascii.buildPacket(message)
        self.assertEqual(expected, actual)
        ModbusRequest.encode = old_encode

    def test_ascii_process_incoming_packets(self):
        """Test ascii process incoming packet."""
        mock_data = b":F7031389000A60\r\n"
        unit = 0x00

        def mock_callback(
            mock_data, *args, **kwargs
        ):  # pylint: disable=unused-argument
            """Mock callback."""

        self._ascii.processIncomingPacket(mock_data, mock_callback, unit)

        # Test failure:
        self._ascii.checkFrame = MagicMock(return_value=False)
        self._ascii.processIncomingPacket(mock_data, mock_callback, unit)

    # ----------------------------------------------------------------------- #
    # Binary tests
    # ----------------------------------------------------------------------- #
    def test_binary_framer_transaction_ready(self):
        """Test a binary frame transaction"""
        msg = TEST_MESSAGE
        self.assertFalse(self._binary.isFrameReady())
        self.assertFalse(self._binary.checkFrame())
        self._binary.addToFrame(msg)
        self.assertTrue(self._binary.isFrameReady())
        self.assertTrue(self._binary.checkFrame())
        self._binary.advanceFrame()
        self.assertFalse(self._binary.isFrameReady())
        self.assertFalse(self._binary.checkFrame())
        self.assertEqual(b"", self._binary.getFrame())

    def test_binary_framer_transaction_full(self):
        """Test a full binary frame transaction"""
        msg = TEST_MESSAGE
        pack = msg[2:-3]
        self._binary.addToFrame(msg)
        self.assertTrue(self._binary.checkFrame())
        result = self._binary.getFrame()
        self.assertEqual(pack, result)
        self._binary.advanceFrame()

    def test_binary_framer_transaction_half(self):
        """Test a half completed binary frame transaction"""
        msg1 = b"\x7b\x01\x03\x00"
        msg2 = b"\x00\x00\x05\x85\xC9\x7d"
        pack = msg1[2:] + msg2[:-3]
        self._binary.addToFrame(msg1)
        self.assertFalse(self._binary.checkFrame())
        result = self._binary.getFrame()
        self.assertEqual(b"", result)
        self._binary.addToFrame(msg2)
        self.assertTrue(self._binary.checkFrame())
        result = self._binary.getFrame()
        self.assertEqual(pack, result)
        self._binary.advanceFrame()

    def test_binary_framer_populate(self):
        """Test a binary frame packet build"""
        request = ModbusRequest()
        self._binary.populateResult(request)
        self.assertEqual(0x00, request.unit_id)

    def test_binary_framer_packet(self):
        """Test a binary frame packet build"""
        old_encode = ModbusRequest.encode
        ModbusRequest.encode = lambda self: b""
        message = ModbusRequest()
        message.unit_id = 0xFF
        message.function_code = 0x01
        expected = b"\x7b\xff\x01\x81\x80\x7d"
        actual = self._binary.buildPacket(message)
        self.assertEqual(expected, actual)
        ModbusRequest.encode = old_encode

    def test_binary_process_incoming_packet(self):
        """Test binary process incoming packet."""
        mock_data = TEST_MESSAGE
        unit = 0x00

        def mock_callback(mock_data):  # pylint: disable=unused-argument
            pass

        self._binary.processIncomingPacket(mock_data, mock_callback, unit)

        # Test failure:
        self._binary.checkFrame = MagicMock(return_value=False)
        self._binary.processIncomingPacket(mock_data, mock_callback, unit)



#!/usr/bin/env python3
"""Pymodbus Asynchronous Client Examples.

The following is an example of how to use the asynchronous modbus
client implementation from pymodbus.
"""
# --------------------------------------------------------------------------- #
# import needed libraries
# --------------------------------------------------------------------------- #
import logging

from twisted.internet import reactor

# from pymodbus.client.asynchronous.udp import AsyncModbusUDPClient
from pymodbus.client.asynchronous import schedulers
from pymodbus.client.asynchronous.tcp import AsyncModbusTCPClient


# --------------------------------------------------------------------------- #
# configure the client logging
# --------------------------------------------------------------------------- #
FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.DEBUG)

# --------------------------------------------------------------------------- #
# helper method to test deferred callbacks
# --------------------------------------------------------------------------- #


def err(*args, **kwargs):
    """Error."""
    txt = f"Err-{args}-{kwargs}"
    logging.error(txt)


def dassert(deferred, callback):
    """Dassert."""

    def _assertor(value):
        assert value  # nosec

    deferred.addCallback(lambda r: _assertor(callback(r)))
    deferred.addErrback(err)


# --------------------------------------------------------------------------- #
# specify slave to query
# --------------------------------------------------------------------------- #
# The slave to query is specified in an optional parameter for each
# individual request. This can be done by specifying the `unit` parameter
# which defaults to `0x00`
# --------------------------------------------------------------------------- #


UNIT = 0x01


def process_response(result):
    """Process response."""
    log.debug(result)


def example_requests(client):
    """Do example requests."""
    rr = client.read_coils(1, 1, unit=0x02)
    rr.addCallback(process_response)
    rr = client.read_holding_registers(1, 1, unit=0x02)
    rr.addCallback(process_response)
    rr = client.read_discrete_inputs(1, 1, unit=0x02)
    rr.addCallback(process_response)
    rr = client.read_input_registers(1, 1, unit=0x02)
    rr.addCallback(process_response)
    stop_asynchronous_test(client)


# --------------------------------------------------------------------------- #
# example requests
# --------------------------------------------------------------------------- #
# simply call the methods that you would like to use. An example session
# is displayed below along with some assert checks. Note that unlike the
# synchronous version of the client, the asynchronous version returns
# deferreds which can be thought of as a handle to the callback to send
# the result of the operation.  We are handling the result using the
# deferred assert helper(dassert).
# --------------------------------------------------------------------------- #


def stop_asynchronous_test(client):
    """Stop async test."""
    # ----------------------------------------------------------------------- #
    # close the client at some time later
    # ----------------------------------------------------------------------- #
    reactor.callLater(1, client.transport.loseConnection)
    reactor.callLater(2, reactor.stop)


def begin_asynchronous_test(client):
    """Begin async test."""
    rq = client.write_coil(1, True, unit=UNIT)
    rr = client.read_coils(1, 1, unit=UNIT)
    dassert(rq, lambda r: not r.isError())  # test for no error
    dassert(rr, lambda r: r.bits[0])  # test the expected value

    rq = client.write_coils(1, [True] * 8, unit=UNIT)
    rr = client.read_coils(1, 8, unit=UNIT)
    dassert(rq, lambda r: not r.isError())  # test for no error
    dassert(rr, lambda r: r.bits == [True] * 8)  # test the expected value

    rq = client.write_coils(1, [False] * 8, unit=UNIT)
    rr = client.read_discrete_inputs(1, 8, unit=UNIT)
    dassert(rq, lambda r: not r.isError())  # test for no error
    dassert(rr, lambda r: r.bits == [True] * 8)  # test the expected value

    rq = client.write_register(1, 10, unit=UNIT)
    rr = client.read_holding_registers(1, 1, unit=UNIT)
    dassert(rq, lambda r: not r.isError())  # test for no error
    dassert(rr, lambda r: r.registers[0] == 10)  # test the expected value

    rq = client.write_registers(1, [10] * 8, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: not r.isError())  # test for no error
    dassert(rr, lambda r: r.registers == [17] * 8)  # test the expected value

    arguments = {
        "read_address": 1,
        "read_count": 8,
        "write_address": 1,
        "write_registers": [20] * 8,
    }
    rq = client.readwrite_registers(arguments, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.registers == [20] * 8)  # test the expected value
    dassert(rr, lambda r: r.registers == [17] * 8)  # test the expected value
    stop_asynchronous_test(client)

    # ----------------------------------------------------------------------- #
    # close the client at some time later
    # ----------------------------------------------------------------------- #
    # reactor.callLater(1, client.transport.loseConnection)
    reactor.callLater(2, reactor.stop)


# --------------------------------------------------------------------------- #
# extra requests
# --------------------------------------------------------------------------- #
# If you are performing a request that is not available in the client
# mixin, you have to perform the request like this instead::
#
# from pymodbus.diag_message import ClearCountersRequest
# from pymodbus.diag_message import ClearCountersResponse
#
# request  = ClearCountersRequest()
# response = client.execute(request)
# if isinstance(response, ClearCountersResponse):
#     ... do something with the response
#
# --------------------------------------------------------------------------- #

# --------------------------------------------------------------------------- #
# choose the client you want
# --------------------------------------------------------------------------- #
# make sure to start an implementation to hit against. For this
# you can use an existing device, the reference implementation in the tools
# directory, or start a pymodbus server.
# --------------------------------------------------------------------------- #


if __name__ == "__main__":
    protocol, deferred = AsyncModbusTCPClient(schedulers.REACTOR, port=5020)
    deferred.addCallback(begin_asynchronous_test)
    deferred.addErrback(err)



"""Test server asyncio."""
import asyncio
from asyncio import CancelledError
import logging
import ssl
import unittest
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pymodbus.datastore import (
    ModbusSequentialDataBlock,
    ModbusServerContext,
    ModbusSlaveContext,
)
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.exceptions import NoSuchSlaveException
from pymodbus.server import (
    StartAsyncTcpServer,
    StartAsyncTlsServer,
    StartAsyncUdpServer,
)


_logger = logging.getLogger()

SERV_IP = "127.0.0.1"
SERV_ADDR = ("127.0.0.1", 0)
TEST_DATA = b"\x01\x00\x00\x00\x00\x06\x01\x03\x00\x00\x00\x01"


class BasicClient(asyncio.BaseProtocol):
    """Basic client."""

    connected = None
    data = None
    dataTo = None
    received_data = None
    done = None
    eof = None
    transport = None
    protocol = None

    def connection_made(self, transport):
        """Get Connection made."""
        _logger.debug("TEST Client connected")
        if BasicClient.connected is not None:
            BasicClient.connected.set_result(True)

        self.transport = transport
        if BasicClient.data is not None:
            _logger.debug("TEST Client write data")
            self.transport.write(BasicClient.data)
        if BasicClient.dataTo is not None:
            _logger.debug("TEST Client sendTo data")
            self.transport.sendto(BasicClient.dataTo)

    def data_received(self, data):
        """Get Data received."""
        _logger.debug("TEST Client data received")
        BasicClient.received_data = data
        if BasicClient.done is not None:
            BasicClient.done.set_result(True)

    def datagram_received(self, data, addr):  # pylint: disable=unused-argument
        """Get Datagram received."""
        _logger.debug("TEST Client datagram received")
        BasicClient.received_data = data
        if BasicClient.done is not None:
            BasicClient.done.set_result(True)
        self.transport.close()

    def connection_lost(self, exc):
        """EOF received."""
        txt = f"TEST Client stream lost: {exc}"
        _logger.debug(txt)
        if BasicClient.eof:
            BasicClient.eof.set_result(True)

    @classmethod
    def clear(cls):
        """Prepare for new round"""
        if BasicClient.transport:
            BasicClient.transport.close()
            BasicClient.transport = None
        BasicClient.data = None
        BasicClient.connected = None
        BasicClient.done = None
        BasicClient.received_data = None
        BasicClient.eof = None
        BasicClient.protocol = None


class AsyncioServerTest(
    unittest.IsolatedAsyncioTestCase
):  # pylint: disable=too-many-public-methods
    """Unittest for the pymodbus.server.asyncio module.

    The scope of this unit test is the life-cycle management of the network
    connections and server objects.

    This unittest suite does not attempt to test any of the underlying protocol details
    """

    def __init__(self, name):
        """Initialize."""
        super().__init__(name)
        self.store = ModbusSlaveContext(
            di=ModbusSequentialDataBlock(0, [17] * 100),
            co=ModbusSequentialDataBlock(0, [17] * 100),
            hr=ModbusSequentialDataBlock(0, [17] * 100),
            ir=ModbusSequentialDataBlock(0, [17] * 100),
        )
        self.context = ModbusServerContext(slaves=self.store, single=True)
        self.identity = ModbusDeviceIdentification(
            info_name={"VendorName": "VendorName"}
        )
        self.server = None
        self.task = None
        self.loop = None

    # -----------------------------------------------------------------------#
    #  Setup/TearDown
    # -----------------------------------------------------------------------#
    def setUp(self):
        """Initialize the test environment by setting up a dummy store and context."""

    async def asyncSetUp(self):
        """Initialize the test environment by setting up a dummy store and context."""
        self.loop = asyncio.get_running_loop()

    async def asyncTearDown(self):
        """Clean up the test environment"""
        if self.server is not None:
            await self.server.server_close()
            self.server = None
        if self.task is not None:
            await asyncio.sleep(0.1)
            if not self.task.cancelled():
                self.task.cancel()
                try:
                    await self.task
                except CancelledError:
                    pass
                self.task = None
        self.context = ModbusServerContext(slaves=self.store, single=True)
        BasicClient.clear()

    def tearDown(self):
        """Clean up the test environment."""

    def handle_task(self, result):
        """Handle task exit."""
        try:
            result = result.result()
        except CancelledError:
            pass

    async def start_server(
        self, do_forever=True, do_defer=True, do_tls=False, do_udp=False, do_ident=False
    ):
        """Handle setup and control of tcp server."""
        args = {
            "context": self.context,
            "address": SERV_ADDR,
            "defer_start": do_defer,
        }
        if do_ident:
            args["identity"] = self.identity
        if do_tls:
            self.server = await StartAsyncTlsServer(**args)
        elif do_udp:
            self.server = await StartAsyncUdpServer(**args)
        else:
            self.server = await StartAsyncTcpServer(**args)
        self.assertIsNotNone(self.server)
        if do_forever:
            self.task = asyncio.create_task(self.server.serve_forever())
            self.task.add_done_callback(self.handle_task)
            self.assertFalse(self.task.cancelled())
            await asyncio.wait_for(self.server.serving, timeout=0.1)
            if not do_udp:
                self.assertIsNotNone(self.server.server)
        elif not do_udp:  # pylint: disable=confusing-consecutive-elif
            self.assertIsNone(self.server.server)
        self.assertEqual(self.server.control.Identity.VendorName, "VendorName")
        await asyncio.sleep(0.1)

    async def connect_server(self):
        """Handle connect to server"""
        BasicClient.connected = self.loop.create_future()
        BasicClient.done = self.loop.create_future()
        BasicClient.eof = self.loop.create_future()
        random_port = self.server.server.sockets[0].getsockname()[
            1
        ]  # get the random server port
        BasicClient.transport, BasicClient.protocol = await self.loop.create_connection(
            BasicClient, host="127.0.0.1", port=random_port
        )
        await asyncio.wait_for(BasicClient.connected, timeout=0.1)
        await asyncio.sleep(0.1)

    # -----------------------------------------------------------------------#
    #  Test ModbusConnectedRequestHandler
    # -----------------------------------------------------------------------#

    async def test_async_start_server_no_loop(self):
        """Test that the modbus tcp asyncio server starts correctly"""
        await self.start_server(do_forever=False)

    async def test_async_start_server(self):
        """Test that the modbus tcp asyncio server starts correctly"""
        await self.start_server()

    async def test_async_tcp_server_serve_forever_twice(self):
        """Call on serve_forever() twice should result in a runtime error"""
        await self.start_server()
        with self.assertRaises(RuntimeError):
            await self.server.serve_forever()

    async def test_async_tcp_server_receive_data(self):
        """Test data sent on socket is received by internals - doesn"t not process data"""
        BasicClient.data = b"\x01\x00\x00\x00\x00\x06\x01\x03\x00\x00\x00\x19"
        await self.start_server()
        with patch(
            "pymodbus.transaction.ModbusSocketFramer.processIncomingPacket",
            new_callable=Mock,
        ) as process:
            await self.connect_server()
            process.assert_called_once()
            self.assertTrue(process.call_args[1]["data"] == BasicClient.data)

    async def test_async_tcp_server_roundtrip(self):
        """Test sending and receiving data on tcp socket"""
        expected_response = b"\x01\x00\x00\x00\x00\x05\x01\x03\x02\x00\x11"
        BasicClient.data = TEST_DATA  # unit 1, read register
        await self.start_server()
        await self.connect_server()
        await asyncio.wait_for(BasicClient.done, timeout=0.1)
        self.assertEqual(BasicClient.received_data, expected_response)

    async def test_async_tcp_server_connection_lost(self):
        """Test tcp stream interruption"""
        await self.start_server()
        await self.connect_server()
        self.assertEqual(len(self.server.active_connections), 1)

        BasicClient.protocol.transport.close()
        await asyncio.sleep(0.2)  # so we have to wait a bit
        self.assertFalse(self.server.active_connections)

    async def test_async_tcp_server_close_active_connection(self):
        """Test server_close() while there are active TCP connections"""
        await self.start_server()
        await self.connect_server()

        # On Windows we seem to need to give this an extra chance to finish,
        # otherwise there ends up being an active connection at the assert.
        await asyncio.sleep(0.5)
        await self.server.server_close()

    async def test_async_tcp_server_no_slave(self):
        """Test unknown slave unit exception"""
        self.context = ModbusServerContext(
            slaves={0x01: self.store, 0x02: self.store}, single=False
        )
        BasicClient.data = b"\x01\x00\x00\x00\x00\x06\x05\x03\x00\x00\x00\x01"
        await self.start_server()
        await self.connect_server()
        self.assertFalse(BasicClient.eof.done())
        self.server.server_close()
        self.server = None

    async def test_async_tcp_server_modbus_error(self):
        """Test sending garbage data on a TCP socket should drop the connection"""
        BasicClient.data = TEST_DATA
        await self.start_server()
        with patch(
            "pymodbus.register_read_message.ReadHoldingRegistersRequest.execute",
            side_effect=NoSuchSlaveException,
        ):
            await self.connect_server()
            await asyncio.wait_for(BasicClient.done, timeout=0.1)

    # -----------------------------------------------------------------------#
    # Test ModbusTlsProtocol
    # -----------------------------------------------------------------------#
    async def test_async_start_tls_server_no_loop(self):
        """Test that the modbus tls asyncio server starts correctly"""
        with patch.object(ssl.SSLContext, "load_cert_chain"):
            await self.start_server(do_tls=True, do_forever=False, do_ident=True)
            self.assertEqual(self.server.control.Identity.VendorName, "VendorName")
            self.assertIsNotNone(self.server.sslctx)

    async def test_async_start_tls_server(self):
        """Test that the modbus tls asyncio server starts correctly"""
        with patch.object(ssl.SSLContext, "load_cert_chain"):
            await self.start_server(do_tls=True, do_ident=True)
            self.assertEqual(self.server.control.Identity.VendorName, "VendorName")
            self.assertIsNotNone(self.server.sslctx)

    async def test_async_tls_server_serve_forever(self):
        """Test StartAsyncTcpServer serve_forever() method"""
        with patch(
            "asyncio.base_events.Server.serve_forever", new_callable=AsyncMock
        ) as serve:
            with patch.object(ssl.SSLContext, "load_cert_chain"):
                await self.start_server(do_tls=True, do_forever=False)
                await self.server.serve_forever()
                serve.assert_awaited()

    async def test_async_tls_server_serve_forever_twice(self):
        """Call on serve_forever() twice should result in a runtime error"""
        with patch.object(ssl.SSLContext, "load_cert_chain"):
            await self.start_server(do_tls=True)
            with pytest.raises(RuntimeError):
                await self.server.serve_forever()

    # -----------------------------------------------------------------------#
    # Test ModbusUdpProtocol
    # -----------------------------------------------------------------------#

    async def test_async_start_udp_server_no_loop(self):
        """Test that the modbus udp asyncio server starts correctly"""
        await self.start_server(do_udp=True, do_forever=False, do_ident=True)
        self.assertEqual(self.server.control.Identity.VendorName, "VendorName")
        self.assertIsNone(self.server.protocol)

    async def test_async_start_udp_server(self):
        """Test that the modbus udp asyncio server starts correctly"""
        await self.start_server(do_udp=True, do_ident=True)
        self.assertEqual(self.server.control.Identity.VendorName, "VendorName")
        self.assertFalse(self.server.protocol is None)

    async def test_async_udp_server_serve_forever_start(self):
        """Test StartAsyncUdpServer serve_forever() method"""
        with patch(
            "asyncio.base_events.Server.serve_forever", new_callable=AsyncMock
        ) as serve:
            await self.start_server(do_forever=False, do_ident=True)
            await self.server.serve_forever()
            serve.assert_awaited()

    async def test_async_udp_server_serve_forever_close(self):
        """Test StarAsyncUdpServer serve_forever() method"""
        await self.start_server(do_udp=True)
        self.assertTrue(asyncio.isfuture(self.server.on_connection_terminated))
        self.assertFalse(self.server.on_connection_terminated.done())

        await self.server.server_close()
        # TBD self.assertTrue(self.server.protocol.is_closing())
        self.server = None

    async def test_async_udp_server_serve_forever_twice(self):
        """Call on serve_forever() twice should result in a runtime error"""
        await self.start_server(do_udp=True, do_ident=True)
        with self.assertRaises(RuntimeError):
            await self.server.serve_forever()

    @pytest.mark.skifif(pytest.IS_WINDOWS, reason="Windows have a timeout problem.")
    async def test_async_udp_server_receive_data(self):
        """Test that the sending data on datagram socket gets data pushed to framer"""
        await self.start_server(do_udp=True)
        with patch(
            "pymodbus.transaction.ModbusSocketFramer.processIncomingPacket",
            new_callable=Mock,
        ) as process:
            self.server.endpoint.datagram_received(data=b"12345", addr=(SERV_IP, 12345))
            await asyncio.sleep(0.1)
            process.seal()
            process.assert_called_once()
            self.assertTrue(process.call_args[1]["data"] == b"12345")

    async def test_async_udp_server_send_data(self):
        """Test that the modbus udp asyncio server correctly sends data outbound"""
        BasicClient.dataTo = b"x\01\x00\x00\x00\x00\x06\x01\x03\x00\x00\x00\x19"
        await self.start_server(do_udp=True)
        random_port = self.server.protocol._sock.getsockname()[  # pylint: disable=protected-access
            1
        ]
        received = self.server.endpoint.datagram_received = Mock(
            wraps=self.server.endpoint.datagram_received
        )
        await self.loop.create_datagram_endpoint(
            BasicClient, remote_addr=("127.0.0.1", random_port)
        )
        await asyncio.sleep(0.1)
        received.assert_called_once()
        self.assertEqual(received.call_args[0][0], BasicClient.dataTo)
        await self.server.server_close()
        self.server = None

    async def test_async_udp_server_roundtrip(self):
        """Test sending and receiving data on udp socket"""
        expected_response = b"\x01\x00\x00\x00\x00\x05\x01\x03\x02\x00\x11"  # value of 17 as per context
        BasicClient.dataTo = TEST_DATA  # unit 1, read register
        BasicClient.done = self.loop.create_future()
        await self.start_server(do_udp=True)
        random_port = self.server.protocol._sock.getsockname()[  # pylint: disable=protected-access
            1
        ]
        transport, _ = await self.loop.create_datagram_endpoint(
            BasicClient, remote_addr=("127.0.0.1", random_port)
        )
        await asyncio.wait_for(BasicClient.done, timeout=0.1)
        self.assertEqual(BasicClient.received_data, expected_response)
        transport.close()

    async def test_async_udp_server_exception(self):
        """Test sending garbage data on a TCP socket should drop the connection"""
        BasicClient.dataTo = b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        BasicClient.connected = self.loop.create_future()
        BasicClient.done = self.loop.create_future()
        await self.start_server(do_udp=True)
        with patch(
            "pymodbus.transaction.ModbusSocketFramer.processIncomingPacket",
            new_callable=lambda: Mock(side_effect=Exception),
        ):
            # get the random server port pylint: disable=protected-access
            random_port = self.server.protocol._sock.getsockname()[1]
            _, _ = await self.loop.create_datagram_endpoint(
                BasicClient, remote_addr=("127.0.0.1", random_port)
            )
            await asyncio.wait_for(BasicClient.connected, timeout=0.1)
            self.assertFalse(BasicClient.done.done())
            self.assertFalse(
                self.server.protocol._sock._closed  # pylint: disable=protected-access
            )

    async def test_async_tcp_server_exception(self):
        """Send garbage data on a TCP socket should drop the connection"""
        BasicClient.data = b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        await self.start_server()
        with patch(
            "pymodbus.transaction.ModbusSocketFramer.processIncomingPacket",
            new_callable=lambda: Mock(side_effect=Exception),
        ):
            await self.connect_server()
            await asyncio.wait_for(BasicClient.eof, timeout=0.1)
            # neither of these should timeout if the test is successful



"""Framer start."""
# pylint: disable=missing-type-doc
from pymodbus.interfaces import IModbusFramer


# Unit ID, Function Code
BYTE_ORDER = ">"
FRAME_HEADER = "BB"

# Transaction Id, Protocol ID, Length, Unit ID, Function Code
SOCKET_FRAME_HEADER = BYTE_ORDER + "HHH" + FRAME_HEADER

# Function Code
TLS_FRAME_HEADER = BYTE_ORDER + "B"


class ModbusFramer(IModbusFramer):
    """Base Framer class."""

    name = ""

    def _validate_unit_id(self, units, single):
        """Validate if the received data is valid for the client.

        :param units: list of unit id for which the transaction is valid
        :param single: Set to true to treat this as a single context
        :return:
        """
        if single:
            return True
        if 0 in units or 0xFF in units:
            # Handle Modbus TCP unit identifier (0x00 0r 0xFF)
            # in asynchronous requests
            return True
        return self._header["uid"] in units  # pylint: disable=no-member

    def sendPacket(self, message):  # pylint: disable=invalid-name
        """Send packets on the bus.

        With 3.5char delay between frames
        :param message: Message to be sent over the bus
        :return:
        """
        return self.client.send(message)  # pylint: disable=no-member

    def recvPacket(self, size):  # pylint: disable=invalid-name
        """Receive packet from the bus.

        With specified len
        :param size: Number of bytes to read
        :return:
        """
        return self.client.recv(size)  # pylint: disable=no-member



"""Test client sync."""
import asyncio
import socket
import ssl

from test.conftest import return_as_coroutine, run_coroutine
from unittest import mock
import pytest

import pymodbus.bit_read_message as pdu_bit_read
import pymodbus.bit_write_message as pdu_bit_write
import pymodbus.client as lib_client
from pymodbus.client.base import ModbusBaseClient, ModbusClientProtocol
from pymodbus.client.mixin import ModbusClientMixin
from pymodbus.constants import Defaults
import pymodbus.diag_message as pdu_diag
from pymodbus.exceptions import ConnectionException, NotImplementedException
from pymodbus.framer.ascii_framer import ModbusAsciiFramer
from pymodbus.framer.rtu_framer import ModbusRtuFramer
from pymodbus.framer.socket_framer import ModbusSocketFramer
from pymodbus.framer.tls_framer import ModbusTlsFramer
import pymodbus.other_message as pdu_other_msg
import pymodbus.register_read_message as pdu_reg_read
import pymodbus.register_write_message as pdu_req_write


@pytest.mark.parametrize(
    "arglist",
    [
        [
            {},
            {"address": 0x01},
            {"address": 0x01, "value": False},
            {"msg": "long message"},
            {"toggle": False},
            {"address": 0x01, "values": [False, True]},
            {"address": 0x01, "values": [22, 44]},
        ]
    ],
)
@pytest.mark.parametrize(
    "method, arg, response",
    [
        ("read_coils", 1, pdu_bit_read.ReadCoilsRequest),
        ("read_discrete_inputs", 1, pdu_bit_read.ReadDiscreteInputsRequest),
        ("read_holding_registers", 1, pdu_reg_read.ReadHoldingRegistersRequest),
        ("read_input_registers", 1, pdu_reg_read.ReadInputRegistersRequest),
        ("write_coil", 2, pdu_bit_write.WriteSingleCoilRequest),
        ("write_register", 2, pdu_req_write.WriteSingleRegisterRequest),
        ("read_exception_status", 0, pdu_other_msg.ReadExceptionStatusRequest),
        ("diag_query_data", 3, pdu_diag.ReturnQueryDataRequest),
        ("diag_restart_communication", 4, pdu_diag.RestartCommunicationsOptionRequest),
        ("diag_read_diagnostic_register", 0, pdu_diag.ReturnDiagnosticRegisterRequest),
        (
            "diag_change_ascii_input_delimeter",
            0,
            pdu_diag.ChangeAsciiInputDelimiterRequest,
        ),
        ("diag_force_listen_only", 0, pdu_diag.ForceListenOnlyModeRequest),
        ("diag_clear_counters", 0, pdu_diag.ClearCountersRequest),
        ("diag_read_bus_message_count", 0, pdu_diag.ReturnBusMessageCountRequest),
        (
            "diag_read_bus_comm_error_count",
            0,
            pdu_diag.ReturnBusCommunicationErrorCountRequest,
        ),
        (
            "diag_read_bus_exception_error_count",
            0,
            pdu_diag.ReturnBusExceptionErrorCountRequest,
        ),
        ("diag_read_slave_message_count", 0, pdu_diag.ReturnSlaveMessageCountRequest),
        (
            "diag_read_slave_no_response_count",
            0,
            pdu_diag.ReturnSlaveNoResponseCountRequest,
        ),
        ("diag_read_slave_nak_count", 0, pdu_diag.ReturnSlaveNAKCountRequest),
        ("diag_read_slave_busy_count", 0, pdu_diag.ReturnSlaveBusyCountRequest),
        (
            "diag_read_bus_char_overrun_count",
            0,
            pdu_diag.ReturnSlaveBusCharacterOverrunCountRequest,
        ),
        ("diag_read_iop_overrun_count", 0, pdu_diag.ReturnIopOverrunCountRequest),
        ("diag_clear_overrun_counter", 0, pdu_diag.ClearOverrunCountRequest),
        ("diag_getclear_modbus_response", 0, pdu_diag.GetClearModbusPlusRequest),
        ("write_coils", 5, pdu_bit_write.WriteMultipleCoilsRequest),
        ("write_registers", 6, pdu_req_write.WriteMultipleRegistersRequest),
        ("readwrite_registers", 1, pdu_reg_read.ReadWriteMultipleRegistersRequest),
        ("mask_write_register", 1, pdu_req_write.MaskWriteRegisterRequest),
    ],
)
def test_client_mixin(arglist, method, arg, response):
    """Test mixin responses."""
    rr = getattr(ModbusClientMixin(), method)(**arglist[arg])
    assert isinstance(rr, response)


@pytest.mark.parametrize(
    "arg_list",
    [
        {
            "fix": {
                "opt_args": {
                    "timeout": Defaults.Timeout + 2,
                    "retries": Defaults.Retries + 2,
                    "retry_on_empty": not Defaults.RetryOnEmpty,
                    "close_comm_on_error": not Defaults.CloseCommOnError,
                    "strict": not Defaults.Strict,
                    "broadcast_enable": not Defaults.BroadcastEnable,
                    "reconnect_delay": 117,
                },
                "defaults": {
                    "timeout": Defaults.Timeout,
                    "retries": Defaults.Retries,
                    "retry_on_empty": Defaults.RetryOnEmpty,
                    "close_comm_on_error": Defaults.CloseCommOnError,
                    "strict": Defaults.Strict,
                    "broadcast_enable": Defaults.BroadcastEnable,
                    "reconnect_delay": Defaults.ReconnectDelay,
                },
            },
            "serial": {
                "pos_arg": "/dev/tty",
                "opt_args": {
                    "framer": ModbusAsciiFramer,
                    "baudrate": Defaults.Baudrate + 500,
                    "bytesize": Defaults.Bytesize - 1,
                    "parity": "E",
                    "stopbits": Defaults.Stopbits + 1,
                    "handle_local_echo": not Defaults.HandleLocalEcho,
                },
                "defaults": {
                    "host": None,
                    "port": "/dev/tty",
                    "framer": ModbusRtuFramer,
                    "baudrate": Defaults.Baudrate,
                    "bytesize": Defaults.Bytesize,
                    "parity": Defaults.Parity,
                    "stopbits": Defaults.Stopbits,
                    "handle_local_echo": Defaults.HandleLocalEcho,
                },
            },
            "tcp": {
                "pos_arg": "192.168.1.2",
                "opt_args": {
                    "port": 112,
                    "framer": ModbusAsciiFramer,
                    "source_address": ("195.6.7.8", 1025),
                },
                "defaults": {
                    "host": "192.168.1.2",
                    "port": Defaults.TcpPort,
                    "framer": ModbusSocketFramer,
                    "source_address": None,
                },
            },
            "tls": {
                "pos_arg": "192.168.1.2",
                "opt_args": {
                    "port": 211,
                    "framer": ModbusAsciiFramer,
                    "source_address": ("195.6.7.8", 1025),
                    "sslctx": None,
                    "certfile": None,
                    "keyfile": None,
                    "password": None,
                },
                "defaults": {
                    "host": "192.168.1.2",
                    "port": Defaults.TlsPort,
                    "framer": ModbusTlsFramer,
                    "source_address": None,
                    "sslctx": None,
                    "certfile": None,
                    "keyfile": None,
                    "password": None,
                },
            },
            "udp": {
                "pos_arg": "192.168.1.2",
                "opt_args": {
                    "port": 121,
                    "framer": ModbusAsciiFramer,
                    "source_address": ("195.6.7.8", 1025),
                },
                "defaults": {
                    "host": "192.168.1.2",
                    "port": Defaults.UdpPort,
                    "framer": ModbusSocketFramer,
                    "source_address": None,
                },
            },
        },
    ],
)
@pytest.mark.parametrize(
    "type_args, clientclass",
    [
        # TBD ("serial", lib_client.AsyncModbusSerialClient),
        # TBD ("serial", lib_client.ModbusSerialClient),
        ("tcp", lib_client.AsyncModbusTcpClient),
        ("tcp", lib_client.ModbusTcpClient),
        ("tls", lib_client.AsyncModbusTlsClient),
        ("tls", lib_client.ModbusTlsClient),
        ("udp", lib_client.AsyncModbusUdpClient),
        ("udp", lib_client.ModbusUdpClient),
    ],
)
@pytest.mark.parametrize("test_default", [True, False])
def test_client_instanciate(
    arg_list,
    type_args,
    clientclass,
    test_default,
):
    """Try to instantiate clients."""
    cur_args = arg_list[type_args]
    if test_default:
        client = clientclass(cur_args["pos_arg"])
        to_test = dict(arg_list["fix"]["defaults"], **cur_args["defaults"])
    else:
        client = clientclass(
            cur_args["pos_arg"],
            **arg_list["fix"]["opt_args"],
            **cur_args["opt_args"],
        )
        to_test = dict(arg_list["fix"]["opt_args"], **cur_args["opt_args"])
        to_test["host"] = cur_args["defaults"]["host"]

    for arg, arg_test in to_test.items():
        assert getattr(client.params, arg) == arg_test

    # Test information methods
    client.last_frame_end = 2
    client.silent_interval = 2
    assert client.idle_time() == 4
    client.last_frame_end = None
    assert not client.idle_time()

    initial_delay = client.delay_ms
    assert initial_delay > 0  # nosec
    client.delay_ms *= 2

    assert client.delay_ms > initial_delay
    client.reset_delay()
    assert client.delay_ms == initial_delay

    assert socket.AF_INET == client._get_address_family(  # pylint: disable=protected-access
        "127.0.0.1"
    )
    assert socket.AF_INET6 == client._get_address_family(  # pylint: disable=protected-access
        "::1"
    )

    # a successful execute
    client.connect = lambda: True
    client.protocol = lambda: True
    client.transaction = mock.Mock(**{"execute.return_value": True})

    # a unsuccessful connect
    client.connect = lambda: False
    client.protocol = None
    with pytest.raises(ConnectionException):
        client.execute()


def test_client_modbusbaseclient():
    """Test modbus base client class."""
    client = ModbusBaseClient(framer=ModbusAsciiFramer)
    client.register(pdu_bit_read.ReadCoilsResponse)
    buffer = "123ABC"
    assert client.send(buffer) == buffer
    assert client.recv(10) == 10

    with pytest.raises(NotImplementedException):
        client.connect()
    with pytest.raises(NotImplementedException):
        client.is_socket_open()
    with pytest.raises(NotImplementedException):
        client.close()

    with mock.patch(
        "pymodbus.client.base.ModbusBaseClient.connect"
    ) as p_connect, mock.patch(
        "pymodbus.client.base.ModbusBaseClient.close"
    ) as p_close:

        p_connect.return_value = True
        p_close.return_value = True
        with ModbusBaseClient(framer=ModbusAsciiFramer) as b_client:
            str(b_client)
        p_connect.return_value = False


async def test_client_made_connection():
    """Test factory protocol made connection."""
    mock_protocol_class = mock.MagicMock()
    client = lib_client.AsyncModbusTcpClient(
        "127.0.0.1", protocol_class=mock_protocol_class
    )
    assert not client.connected
    assert client.protocol is None
    client.protocol_made_connection(mock.sentinel.PROTOCOL)
    assert client.connected
    assert client.protocol is mock.sentinel.PROTOCOL

    client.protocol_made_connection(mock.sentinel.PROTOCOL_UNEXPECTED)
    assert client.connected
    assert client.protocol is mock.sentinel.PROTOCOL


async def test_client_lost_connection():
    """Test factory protocol lost connection."""
    mock_protocol_class = mock.MagicMock()
    client = lib_client.AsyncModbusTcpClient(
        "127.0.0.1", protocol_class=mock_protocol_class
    )
    assert not client.connected
    assert client.protocol is None

    # fake client is connected and *then* looses connection:
    client.connected = True
    client.params.host = mock.sentinel.HOST
    client.params.port = mock.sentinel.PORT
    client.protocol = mock.sentinel.PROTOCOL
    client.protocol_lost_connection(mock.sentinel.PROTOCOL_UNEXPECTED)
    assert not client.connected

    client.connected = True
    with mock.patch(
        "pymodbus.client.tcp.AsyncModbusTcpClient._reconnect"
    ) as mock_reconnect:
        mock_reconnect.return_value = mock.sentinel.RECONNECT_GENERATOR

        client.protocol_lost_connection(mock.sentinel.PROTOCOL)
    assert not client.connected
    assert client.protocol is None


async def test_client_base_async():
    """Test modbus base client class."""
    with mock.patch(
        "pymodbus.client.base.ModbusBaseClient.connect"
    ) as p_connect, mock.patch(
        "pymodbus.client.base.ModbusBaseClient.close"
    ) as p_close:

        loop = asyncio.get_event_loop()
        p_connect.return_value = loop.create_future()
        p_connect.return_value.set_result(True)
        p_close.return_value = loop.create_future()
        p_close.return_value.set_result(True)
        async with ModbusBaseClient(framer=ModbusAsciiFramer) as client:
            str(client)
        p_connect.return_value = loop.create_future()
        p_connect.return_value.set_result(False)
        p_close.return_value = loop.create_future()
        p_close.return_value.set_result(False)


async def test_client_protocol():
    """Test base modbus async client protocol."""
    protocol = ModbusClientProtocol(framer=ModbusSocketFramer)
    assert protocol.factory is None
    assert protocol.transport is None
    assert not protocol.connected

    protocol.factory = mock.MagicMock()
    protocol.connection_made(mock.sentinel.TRANSPORT)
    assert protocol.transport is mock.sentinel.TRANSPORT
    protocol.factory.protocol_made_connection.assert_called_once_with(  # pylint: disable=no-member
        protocol
    )
    assert (
        not protocol.factory.protocol_lost_connection.call_count  # nosec pylint: disable=no-member
    )

    protocol.factory.reset_mock()
    protocol.connection_lost(mock.sentinel.REASON)
    assert protocol.transport is None  # nosec
    assert (
        not protocol.factory.protocol_made_connection.call_count  # nosec pylint: disable=no-member
    )
    protocol.factory.protocol_lost_connection.assert_called_once_with(  # pylint: disable=no-member
        protocol
    )
    protocol.raise_future = mock.MagicMock()
    request = mock.MagicMock()
    protocol.transaction.addTransaction(request, 1)
    protocol.connection_lost(mock.sentinel.REASON)
    protocol.raise_future.assert_called_once()
    call_args = protocol.raise_future.call_args.args
    assert call_args[0] == request
    assert isinstance(call_args[1], ConnectionException)
    protocol.transport = mock.MagicMock()
    await protocol.close()


async def test_client_protocol_receiver():
    """Test the client protocol data received"""
    protocol = ModbusClientProtocol(framer=ModbusSocketFramer)
    transport = mock.MagicMock()
    protocol.connection_made(transport)
    assert protocol.transport == transport
    assert protocol.connected
    data = b"\x00\x00\x12\x34\x00\x06\xff\x01\x01\x02\x00\x04"

    # setup existing request
    assert not list(protocol.transaction)
    response = protocol._build_response(0x00)  # pylint: disable=protected-access
    protocol.data_received(data)
    result = response.result()
    assert isinstance(result, pdu_bit_read.ReadCoilsResponse)

    protocol._connected = False  # pylint: disable=protected-access
    with pytest.raises(ConnectionException):
        await protocol._build_response(0x00)  # pylint: disable=protected-access


async def test_client_protocol_response():
    """Test the udp client protocol builds responses"""
    protocol = ModbusClientProtocol(framer=ModbusSocketFramer)
    response = protocol._build_response(0x00)  # pylint: disable=protected-access
    excp = response.exception()
    assert isinstance(excp, ConnectionException)
    assert not list(protocol.transaction)

    protocol._connected = True  # pylint: disable=protected-access
    protocol._build_response(0x00)  # pylint: disable=protected-access
    assert len(list(protocol.transaction)) == 1


async def test_client_protocol_handler():
    """Test the client protocol handles responses"""
    protocol = ModbusClientProtocol(framer=ModbusSocketFramer)
    transport = mock.MagicMock()
    protocol.connection_made(transport=transport)
    reply = pdu_bit_read.ReadCoilsRequest(1, 1)
    reply.transaction_id = 0x00
    protocol._handle_response(None)  # pylint: disable=protected-access
    protocol._handle_response(reply)  # pylint: disable=protected-access
    response = protocol._build_response(0x00)  # pylint: disable=protected-access
    protocol._handle_response(reply)  # pylint: disable=protected-access
    result = response.result()
    assert result == reply


async def test_client_protocol_execute():
    """Test the client protocol execute method"""
    protocol = ModbusClientProtocol("127.0.0.1", framer=ModbusSocketFramer)
    protocol.create_future = mock.MagicMock()
    fut = asyncio.Future()
    fut.set_result(fut)
    protocol.create_future.return_value = fut
    transport = mock.MagicMock()
    protocol.connection_made(transport)
    protocol.transport.write = mock.Mock()

    request = pdu_bit_read.ReadCoilsRequest(1, 1)
    response = await protocol.execute(request)
    tid = request.transaction_id
    f_trans = protocol.transaction.getTransaction(tid)
    assert response == f_trans

    protocol.params.broadcast_enable = True
    request = pdu_bit_read.ReadCoilsRequest(1, 1)
    response = await protocol.execute(request)


def test_client_udp():
    """Test client udp."""
    protocol = ModbusClientProtocol("127.0.0.1", framer=ModbusSocketFramer)
    protocol.datagram_received(bytes("00010000", "utf-8"), 1)
    protocol.transport = mock.MagicMock()
    protocol.use_udp = True
    protocol.write_transport(bytes("00010000", "utf-8"))


def test_client_udp_connect():
    """Test the Udp client connection method"""
    with mock.patch.object(socket, "socket") as mock_method:

        class DummySocket:  # pylint: disable=too-few-public-methods
            """Dummy socket."""

            def settimeout(self, *a, **kwa):
                """Set timeout."""

        mock_method.return_value = DummySocket()
        client = lib_client.ModbusUdpClient("127.0.0.1")
        assert client.connect()

    with mock.patch.object(socket, "socket") as mock_method:
        mock_method.side_effect = socket.error()
        client = lib_client.ModbusUdpClient("127.0.0.1")
        assert not client.connect()


def test_client_tcp_connect():
    """Test the tcp client connection method"""
    with mock.patch.object(socket, "create_connection") as mock_method:
        _socket = mock.MagicMock()
        mock_method.return_value = _socket
        client = lib_client.ModbusTcpClient("127.0.0.1")
        _socket.getsockname.return_value = ("dmmy", 1234)
        assert client.connect()

    with mock.patch.object(socket, "create_connection") as mock_method:
        mock_method.side_effect = socket.error()
        client = lib_client.ModbusTcpClient("127.0.0.1")
        assert not client.connect()


def test_client_tls_connect():
    """Test the tls client connection method"""
    with mock.patch.object(ssl.SSLSocket, "connect") as mock_method:
        client = lib_client.ModbusTlsClient("127.0.0.1")
        assert client.connect()

    with mock.patch.object(socket, "create_connection") as mock_method:
        mock_method.side_effect = socket.error()
        client = lib_client.ModbusTlsClient("127.0.0.1")
        assert not client.connect()


@mock.patch("pymodbus.client.tcp.asyncio.sleep")
async def test_client_reconnect(mock_sleep):
    """Test factory reconnect."""
    mock_protocol_class = mock.MagicMock()
    mock_sleep.side_effect = return_as_coroutine()
    loop = asyncio.get_running_loop()
    loop.create_connection = mock.MagicMock(return_value=(None, None))
    client = lib_client.AsyncModbusTcpClient(
        "127.0.0.1", protocol_class=mock_protocol_class
    )
    client.delay_ms = 5000
    await client.connect()

    run_coroutine(client._reconnect())  # pylint: disable=protected-access
    mock_sleep.assert_called_once_with(5)
    assert loop.create_connection.call_count >= 1  # nosec



"""Configure pytest."""
import functools
import platform

import pytest

from pymodbus.interfaces import IModbusSlaveContext


def pytest_configure():
    """Configure pytest."""
    pytest.IS_DARWIN = platform.system().lower() == "darwin"
    pytest.IS_WINDOWS = platform.system().lower() == "windows"


# -----------------------------------------------------------------------#
# Generic fixtures
# -----------------------------------------------------------------------#


class MockContext(IModbusSlaveContext):
    """Mock context."""

    def __init__(self, valid=False, default=True):
        """Initialize."""
        self.valid = valid
        self.default = default

    def validate(self, fx, address, count=0):
        """Validate values."""
        return self.valid

    def getValues(self, fx, address, count=0):
        """Get values."""
        return [self.default] * count

    def setValues(self, fx, address, values):
        """Set values."""


class MockLastValuesContext(IModbusSlaveContext):
    """Mock context."""

    def __init__(self, valid=False, default=True):
        """Initialize."""
        self.valid = valid
        self.default = default
        self.last_values = []

    def validate(self, fx, address, count=0):
        """Validate values."""
        return self.valid

    def getValues(self, fx, address, count=0):
        """Get values."""
        return [self.default] * count

    def setValues(self, fx, address, values):
        """Set values."""
        self.last_values = values


class FakeList:
    """Todo, replace with magic mock."""

    def __init__(self, size):
        """Initialize."""
        self.size = size

    def __len__(self):
        """Get length."""
        return self.size

    def __iter__(self):
        """Iterate."""


class mockSocket:  # pylint: disable=invalid-name
    """Mock socket."""

    timeout = 2

    def __init__(self):
        """Initialize."""
        self.data = None

    def mock_store(self, msg):
        """Store message."""
        self.data = msg

    def mock_retrieve(self, size):
        """Get message."""
        if not self.data or not size:
            return b""
        if size >= len(self.data):
            retval = self.data
        else:
            retval = self.data[0:size]
        self.data = None
        return retval

    def close(self):
        """Close."""
        return True

    def recv(self, size):
        """Receive."""
        return self.mock_retrieve(size)

    def read(self, size):
        """Read."""
        return self.mock_retrieve(size)

    def send(self, msg):
        """Send."""
        self.mock_store(msg)
        return len(msg)

    def recvfrom(self, size):
        """Receive from."""
        return [self.mock_retrieve(size)]

    def sendto(self, msg, *args):  # pylint: disable=unused-argument
        """Send to."""
        self.mock_store(msg)
        return len(msg)

    def setblocking(self, flag):  # pylint: disable=unused-argument
        """Set blocking."""
        return None


def run_coroutine(coro):
    """Run a coroutine as top-level task by iterating through all yielded steps."""
    result = None
    try:
        # step through all parts of coro without scheduling anything else:
        while True:
            result = coro.send(result)
    except StopIteration as exc:
        # coro reached end pass on its return value:
        return exc.value


def _yielded_return(return_value, *args):  # pylint: disable=unused-argument
    """Return Generator factory function with return value."""

    async def _():
        """Actual generator producing value."""
        # yield

    # return new generator each time this function is called:
    return _()


def return_as_coroutine(return_value=None):
    """Create a function that behaves like an asyncio coroutine and returns the given value.

    Typically used as a side effect of a mocked coroutine like this:

        # in module mymod:
        @asyncio.coroutine
        def my_coro_under_test():
            yield from asyncio.sleep(1)
            yield from asyncio.sleep(2)
            return 42

        # in test module:
        @mock.patch("mymod.asyncio.sleep")
        def test_it(mock_sleep):
            mock_sleep.side_effect = return_as_coroutine()
            result = run_coroutine(my_coro_under_test)
    """
    return functools.partial(_yielded_return, return_value)



#!/usr/bin/env python3
"""Installs pymodbus using setuptools."""


# --------------------------------------------------------------------------- #
# initialization
# --------------------------------------------------------------------------- #
from setuptools import find_packages, setup
try:
    from setup_commands import command_classes
except ImportError:
    command_classes = {}
from pymodbus import __author__, __maintainer__, __version__

dependencies = {}
with open('requirements.txt') as reqs:
    option = None
    for line in reqs.read().split('\n'):
        if line == '':
            option = None
        elif line.startswith('# install:'):
            option = line.split(':')[1]
            dependencies[option] = []
        elif not line.startswith('#') and option:
            dependencies[option].append(line)

install_req = dependencies['required']
del dependencies['required']


# --------------------------------------------------------------------------- #
# configuration
# --------------------------------------------------------------------------- #
setup(
    install_requires=install_req,
    extras_require=dependencies,
    cmdclass=command_classes,
)



#!/usr/bin/env python3
"""Pymodbus Synchronous Server Example with Custom functions.

Implements a custom function code not in standard modbus function code list
and its response which otherwise would throw  `IllegalFunction (0x1)` error.

Steps:
1. Create CustomModbusRequest class derived from ModbusRequest
    ```class CustomModbusRequest(ModbusRequest):
           function_code = 75   # Value less than 0x80)
           _rtu_frame_size = <some_value>  # Required only For RTU support

           def __init__(custom_arg=None, **kwargs)
                # Make sure the arguments has default values, will error out
                # while decoding otherwise
                ModbusRequest.__init__(self, **kwargs)
                self.custom_request_arg = custom_arg

            def encode(self):
                # Implement encoding logic

            def decode(self, data):
                # implement decoding logic

            def execute(self, context):
                # Implement execute logic
                ...
                return CustomModbusResponse(..)

    ```
2. Create CustomModbusResponse class derived from ModbusResponse
    ```class CustomModbusResponse(ModbusResponse):
           function_code = 75   # Value less than 0x80)
           _rtu_byte_count_pos = <some_value>  # Required only For RTU support

           def __init__(self, custom_args=None, **kwargs):
                # Make sure the arguments has default values, will error out
                # while decoding otherwise
                ModbusResponse.__init__(self, **kwargs)
                self.custom_reponse_values = values

            def encode(self):
                # Implement encoding logic
            def decode(self, data):
                # Implement decoding logic
    ```
3. Register with ModbusSlaveContext,
    if your request has to access some values from the data-store.
    ```store = ModbusSlaveContext(...)
       store.register(CustomModbusRequest.function_code, "dummy_context_name")
    ```
4. Pass CustomModbusRequest class as argument to Start<protocol>Server
    ```
    StartTcpServer(..., custom_functions=[CustomModbusRequest]..)
    ```

"""
import logging

from .custom_message import (  # pylint: disable=relative-beyond-top-level
    CustomModbusRequest,
)

from pymodbus.datastore import (
    ModbusSequentialDataBlock,
    ModbusServerContext,
    ModbusSlaveContext,
)
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.server import StartTcpServer
from pymodbus.version import version


# --------------------------------------------------------------------------- #
# configure the service logging
# --------------------------------------------------------------------------- #

FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.DEBUG)


def run_server():
    """Run server."""
    store = ModbusSlaveContext(
        di=ModbusSequentialDataBlock(0, [17] * 100),
        co=ModbusSequentialDataBlock(0, [17] * 100),
        hr=ModbusSequentialDataBlock(0, [17] * 100),
        ir=ModbusSequentialDataBlock(0, [17] * 100),
    )

    store.register(
        CustomModbusRequest.function_code,
        "cm",
        ModbusSequentialDataBlock(0, [17] * 100),
    )
    context = ModbusServerContext(slaves=store, single=True)

    # ----------------------------------------------------------------------- #
    # initialize the server information
    # ----------------------------------------------------------------------- #
    # If you don"t set this or any fields, they are defaulted to empty strings.
    # ----------------------------------------------------------------------- #
    identity = ModbusDeviceIdentification(
        info_name={
            "VendorName": "Pymodbus",
            "ProductCode": "PM",
            "VendorUrl": "https://github.com/riptideio/pymodbus/",
            "ProductName": "Pymodbus Server",
            "ModelName": "Pymodbus Server",
            "MajorMinorRevision": version.short(),
        }
    )

    # ----------------------------------------------------------------------- #
    # run the server you want
    # ----------------------------------------------------------------------- #
    # Tcp:
    StartTcpServer(
        context,
        identity=identity,
        address=("localhost", 5020),
        custom_functions=[CustomModbusRequest],
    )


if __name__ == "__main__":
    run_server()



"""Test pdu."""
import unittest

from pymodbus.exceptions import NotImplementedException
from pymodbus.pdu import (
    ExceptionResponse,
    IllegalFunctionRequest,
    ModbusExceptions,
    ModbusRequest,
    ModbusResponse,
)


class SimplePduTest(unittest.TestCase):
    """Unittest for the pymod.pdu module."""

    def setUp(self):
        """Initialize the test environment"""
        self.bad_requests = (
            ModbusRequest(),
            ModbusResponse(),
        )
        self.illegal = IllegalFunctionRequest(1)
        self.exception = ExceptionResponse(1, 1)

    def tearDown(self):
        """Clean up the test environment"""
        del self.bad_requests
        del self.illegal
        del self.exception

    def test_not_impelmented(self):
        """Test a base classes for not implemented functions"""
        for request in self.bad_requests:
            self.assertRaises(NotImplementedException, request.encode)

        for request in self.bad_requests:
            self.assertRaises(NotImplementedException, request.decode, None)

    def test_error_methods(self):
        """Test all error methods"""
        self.illegal.decode("12345")
        self.illegal.execute(None)

        result = self.exception.encode()
        self.exception.decode(result)
        self.assertEqual(result, b"\x01")
        self.assertEqual(self.exception.exception_code, 1)

    def test_request_exception_factory(self):
        """Test all error methods"""
        request = ModbusRequest()
        request.function_code = 1
        errors = {ModbusExceptions.decode(c): c for c in range(1, 20)}
        for error, code in iter(errors.items()):
            result = request.doException(code)
            self.assertEqual(str(result), f"Exception Response(129, 1, {error})")

    def test_calculate_rtu_frame_size(self):
        """Test the calculation of Modbus/RTU frame sizes"""
        self.assertRaises(
            NotImplementedException, ModbusRequest.calculateRtuFrameSize, b""
        )
        ModbusRequest._rtu_frame_size = 5  # pylint: disable=protected-access
        self.assertEqual(ModbusRequest.calculateRtuFrameSize(b""), 5)
        del ModbusRequest._rtu_frame_size

        ModbusRequest._rtu_byte_count_pos = 2  # pylint: disable=protected-access
        self.assertEqual(
            ModbusRequest.calculateRtuFrameSize(
                b"\x11\x01\x05\xcd\x6b\xb2\x0e\x1b\x45\xe6"
            ),
            0x05 + 5,
        )
        del ModbusRequest._rtu_byte_count_pos

        self.assertRaises(
            NotImplementedException, ModbusResponse.calculateRtuFrameSize, b""
        )
        ModbusResponse._rtu_frame_size = 12  # pylint: disable=protected-access
        self.assertEqual(ModbusResponse.calculateRtuFrameSize(b""), 12)
        del ModbusResponse._rtu_frame_size
        ModbusResponse._rtu_byte_count_pos = 2  # pylint: disable=protected-access
        self.assertEqual(
            ModbusResponse.calculateRtuFrameSize(
                b"\x11\x01\x05\xcd\x6b\xb2\x0e\x1b\x45\xe6"
            ),
            0x05 + 5,
        )
        del ModbusResponse._rtu_byte_count_pos



#!/usr/bin/env python3
"""Pymodbus Synchronous Client standard calls example.

This example uses client_sync.py to handle connection, and have the same options.

The corresponding server must be started before e.g. as:
    python3 server_sync.py
"""
import logging

from examples.client_sync import run_sync_client, setup_sync_client


SLAVE = 0x01


def _handle_coils(client):
    """Read/Write coils."""
    _logger.info("### Reading Coil")
    rr = client.read_coils(1, 1, slave=SLAVE)
    assert not rr.isError()  # test that call was OK
    txt = f"### coils response: {str(rr.bits)}"
    _logger.debug(txt)

    _logger.info("### Reading Coils to get bit 5")
    rr = client.read_coils(1, 5, slave=SLAVE)
    assert not rr.isError()  # test that call was OK
    txt = f"### coils response: {str(rr.bits)}"
    _logger.debug(txt)

    _logger.info("### Write true to coil bit 0 and read to verify")
    rq = client.write_coil(0, True, slave=SLAVE)
    rr = client.read_coils(0, 1, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    assert rr.bits[0]  # test the expected value
    txt = f"### coils response: {str(rr.bits)}"
    _logger.debug(txt)

    _logger.info("### Write true to multiple coils 1-8")
    rq = client.write_coils(1, [True] * 21, slave=SLAVE)
    rr = client.read_coils(1, 21, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    resp = [True] * 21
    # If the returned output quantity is not a multiple of eight,
    # the remaining bits in the final data byte will be padded with zeros
    # (toward the high order end of the byte).
    resp.extend([False] * 3)
    assert rr.bits == resp  # test the expected value
    txt = f"### coils response: {str(rr.bits)}"
    _logger.debug(txt)

    _logger.info("### Write False to address 1-8 coils")
    rq = client.write_coils(1, [False] * 8, slave=SLAVE)
    rr = client.read_coils(1, 8, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    assert rr.bits == [False] * 8  # test the expected value
    txt = f"### coils response: {str(rr.bits)}"
    _logger.debug(txt)


def _handle_discrete_input(client):
    """Read discrete inputs."""
    _logger.info("### Reading discrete input, Read address:0-7")
    rr = client.read_discrete_inputs(0, 8, slave=SLAVE)
    assert not rr.isError()  # nosec test that we are not an error
    txt = f"### address 0-7 is: {str(rr.bits)}"
    _logger.debug(txt)


def _handle_holding_registers(client):
    """Read/write holding registers."""
    _logger.info("### write holding register and read holding registers")
    rq = client.write_register(1, 10, slave=SLAVE)
    rr = client.read_holding_registers(1, 1, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    assert rr.registers[0] == 10  # nosec test the expected value
    txt = f"### address 1 is: {str(rr.registers[0])}"
    _logger.debug(txt)

    _logger.info("### write holding registers and read holding registers")
    rq = client.write_registers(1, [10] * 8, slave=SLAVE)
    rr = client.read_holding_registers(1, 8, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    assert rr.registers == [10] * 8  # nosec test the expected value
    txt = f"### address 1-8 is: {str(rr.registers)}"
    _logger.debug(txt)

    _logger.info("### write read holding registers")
    arguments = {
        "read_address": 1,
        "read_count": 8,
        "write_address": 1,
        "write_registers": [256, 128, 100, 50, 25, 10, 5, 1],
    }
    rq = client.readwrite_registers(unit=SLAVE, **arguments)
    rr = client.read_holding_registers(1, 8, slave=SLAVE)
    assert not rq.isError() and not rr.isError()  # test that calls was OK
    assert rq.registers == arguments["write_registers"]
    assert rr.registers == arguments["write_registers"]
    txt = f"### Test 8 read result: address 1-8 is: {str(rr.registers)}"
    _logger.debug(txt)


def _handle_input_registers(client):
    """Read input registers."""
    _logger.info("### read input registers")
    rr = client.read_input_registers(1, 8, slave=SLAVE)
    assert not rr.isError()  # nosec test that we are not an error
    txt = f"### address 1 is: {str(rr.registers[0])}"
    _logger.debug(txt)


def run_sync_basic_calls(client):
    """Demonstrate basic read/write calls."""
    _handle_coils(client)
    _handle_discrete_input(client)
    _handle_holding_registers(client)
    _handle_input_registers(client)


# --------------------------------------------------------------------------- #
# Extra code, to allow commandline parameters instead of changing the code
# --------------------------------------------------------------------------- #
FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
logging.basicConfig(format=FORMAT)
_logger = logging.getLogger()


if __name__ == "__main__":
    testclient = setup_sync_client()
    run_sync_client(testclient, modbus_calls=run_sync_basic_calls)



"""Modbus Clients to be used with REPL."""
# pylint: disable=missing-type-doc
import functools

from pymodbus.client import (
    ModbusSerialClient as _ModbusSerialClient,
    ModbusTcpClient as _ModbusTcpClient,
)
from pymodbus.constants import Defaults
from pymodbus.diag_message import (
    ChangeAsciiInputDelimiterRequest,
    ClearCountersRequest,
    ClearOverrunCountRequest,
    ForceListenOnlyModeRequest,
    GetClearModbusPlusRequest,
    RestartCommunicationsOptionRequest,
    ReturnBusCommunicationErrorCountRequest,
    ReturnBusExceptionErrorCountRequest,
    ReturnBusMessageCountRequest,
    ReturnDiagnosticRegisterRequest,
    ReturnIopOverrunCountRequest,
    ReturnQueryDataRequest,
    ReturnSlaveBusCharacterOverrunCountRequest,
    ReturnSlaveBusyCountRequest,
    ReturnSlaveMessageCountRequest,
    ReturnSlaveNAKCountRequest,
    ReturnSlaveNoResponseCountRequest,
)
from pymodbus.exceptions import ModbusIOException
from pymodbus.mei_message import ReadDeviceInformationRequest
from pymodbus.other_message import (
    GetCommEventCounterRequest,
    GetCommEventLogRequest,
    ReadExceptionStatusRequest,
    ReportSlaveIdRequest,
)
from pymodbus.pdu import ExceptionResponse, ModbusExceptions


def make_response_dict(resp):
    """Make response dict."""
    resp_dict = {"function_code": resp.function_code, "address": resp.address}
    if hasattr(resp, "value"):
        resp_dict["value"] = resp.value
    elif hasattr(resp, "values"):
        resp_dict["values"] = resp.values
    elif hasattr(resp, "count"):
        resp_dict["count"] = resp.count
    return resp_dict


def handle_brodcast(func):
    """Handle broadcast."""

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        self = args[0]
        resp = func(*args, **kwargs)
        if not kwargs.get("unit") and self.params.broadcast_enable:
            return {"broadcasted": True}
        if not resp.isError():
            return make_response_dict(resp)
        return ExtendedRequestSupport._process_exception(  # pylint: disable=protected-access
            resp, **kwargs
        )

    return _wrapper


class ExtendedRequestSupport:  # pylint: disable=(too-many-public-methods
    """Extended request support."""

    @staticmethod
    def _process_exception(resp, **kwargs):
        """Set internal process exception."""
        unit = kwargs.get("unit")
        if unit == 0:  # pylint: disable=compare-to-zero,disable=consider-using-assignment-expr
            err = {"message": "Broadcast message, ignoring errors!!!"}
        else:
            if isinstance(resp, ExceptionResponse):  # pylint: disable=else-if-used
                err = {
                    "original_function_code": f"{resp.original_code} ({hex(resp.original_code)})",
                    "error_function_code": f"{resp.function_code} ({hex(resp.function_code)})",
                    "exception code": resp.exception_code,
                    "message": ModbusExceptions.decode(resp.exception_code),
                }
            elif isinstance(resp, ModbusIOException):
                err = {
                    "original_function_code": f"{resp.fcode} ({hex(resp.fcode)})",
                    "error": resp.message,
                }
            else:
                err = {"error": str(resp)}
        return err

    def read_coils(self, address, count=1, slave=Defaults.Slave, **kwargs):
        """Read `count` coils from a given slave starting at `address`.

        :param address: The starting address to read from
        :param count: The number of coils to read
        :param slave: Modbus slave unit ID
        :param kwargs:
        :returns: List of register values
        """
        resp = super().read_coils(  # pylint: disable=no-member
            address, count, slave, **kwargs
        )
        if not resp.isError():
            return {"function_code": resp.function_code, "bits": resp.bits}
        return ExtendedRequestSupport._process_exception(resp)

    def read_discrete_inputs(self, address, count=1, slave=Defaults.Slave, **kwargs):
        """Read `count` number of discrete inputs starting at offset `address`.

        :param address: The starting address to read from
        :param count: The number of coils to read
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return: List of bits
        """
        resp = super().read_discrete_inputs(  # pylint: disable=no-member
            address, count, slave, **kwargs
        )
        if not resp.isError():
            return {"function_code": resp.function_code, "bits": resp.bits}
        return ExtendedRequestSupport._process_exception(resp)

    @handle_brodcast
    def write_coil(self, address, value, slave=Defaults.Slave, **kwargs):
        """Write `value` to coil at `address`.

        :param address: coil offset to write to
        :param value: bit value to write
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().write_coil(  # pylint: disable=no-member
            address, value, slave, **kwargs
        )
        return resp

    @handle_brodcast
    def write_coils(self, address, values, slave=Defaults.Slave, **kwargs):
        """Write `value` to coil at `address`.

        :param address: coil offset to write to
        :param values: list of bit values to write (comma separated)
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().write_coils(  # pylint: disable=no-member
            address, values, slave, **kwargs
        )
        return resp

    @handle_brodcast
    def write_register(self, address, value, slave=Defaults.Slave, **kwargs):
        """Write `value` to register at `address`.

        :param address: register offset to write to
        :param value: register value to write
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().write_register(  # pylint: disable=no-member
            address, value, slave, **kwargs
        )
        return resp

    @handle_brodcast
    def write_registers(self, address, values, slave=Defaults.Slave, **kwargs):
        """Write list of `values` to registers starting at `address`.

        :param address: register offset to write to
        :param values: list of register value to write (comma separated)
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().write_registers(  # pylint: disable=no-member
            address, values, slave, **kwargs
        )
        return resp

    def read_holding_registers(self, address, count=1, slave=Defaults.Slave, **kwargs):
        """Read `count` number of holding registers starting at `address`.

        :param address: starting register offset to read from
        :param count: Number of registers to read
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().read_holding_registers(  # pylint: disable=no-member
            address, count, slave, **kwargs
        )
        if not resp.isError():
            return {"function_code": resp.function_code, "registers": resp.registers}
        return ExtendedRequestSupport._process_exception(resp)

    def read_input_registers(self, address, count=1, slave=Defaults.Slave, **kwargs):
        """Read `count` number of input registers starting at `address`.

        :param address: starting register offset to read from to
        :param count: Number of registers to read
        :param slave: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().read_input_registers(  # pylint: disable=no-member
            address, count, slave, **kwargs
        )
        if not resp.isError():
            return {"function_code": resp.function_code, "registers": resp.registers}
        return ExtendedRequestSupport._process_exception(resp)

    def readwrite_registers(
        self,
        read_address,
        read_count,
        write_address,
        write_registers,
        unit=Defaults.Slave,
        **kwargs,
    ):
        """Read `read_count` number of holding registers.

        Starting at `read_address`
        and write `write_registers` starting at `write_address`.

        :param read_address: register offset to read from
        :param read_count: Number of registers to read
        :param write_address: register offset to write to
        :param write_registers: List of register values to write (comma separated)
        :param unit: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().readwrite_registers(  # pylint: disable=no-member
            read_address=read_address,
            read_count=read_count,
            write_address=write_address,
            write_registers=write_registers,
            unit=unit,
            **kwargs,
        )
        if not resp.isError():
            return {"function_code": resp.function_code, "registers": resp.registers}
        return ExtendedRequestSupport._process_exception(resp)

    def mask_write_register(
        self,
        address=0x0000,
        and_mask=0xFFFF,
        or_mask=0x0000,
        unit=Defaults.Slave,
        **kwargs,
    ):
        """Mask content of holding register at `address` with `and_mask` and `or_mask`.

        :param address: Reference address of register
        :param and_mask: And Mask
        :param or_mask: OR Mask
        :param unit: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        resp = super().mask_write_register(  # pylint: disable=no-member
            address=address, and_mask=and_mask, or_mask=or_mask, unit=unit, **kwargs
        )
        if not resp.isError():
            return {
                "function_code": resp.function_code,
                "address": resp.address,
                "and mask": resp.and_mask,
                "or mask": resp.or_mask,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def read_device_information(self, read_code=None, object_id=0x00, **kwargs):
        """Read the identification and additional information of remote slave.

        :param read_code:  Read Device ID code (0x01/0x02/0x03/0x04)
        :param object_id: Identification of the first object to obtain.
        :param kwargs:
        :return:
        """
        request = ReadDeviceInformationRequest(read_code, object_id, **kwargs)
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {
                "function_code": resp.function_code,
                "information": resp.information,
                "object count": resp.number_of_objects,
                "conformity": resp.conformity,
                "next object id": resp.next_object_id,
                "more follows": resp.more_follows,
                "space left": resp.space_left,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def report_slave_id(self, unit=Defaults.Slave, **kwargs):
        """Report information about remote slave ID.

        :param unit: Modbus slave unit ID
        :param kwargs:
        :return:
        """
        request = ReportSlaveIdRequest(unit, **kwargs)
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {
                "function_code": resp.function_code,
                "identifier": resp.identifier.decode("cp1252"),
                "status": resp.status,
                "byte count": resp.byte_count,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def read_exception_status(self, **kwargs):
        """Read tcontents of eight Exception Status output.

        In a remote device.

        :param kwargs:
        :return:
        """
        request = ReadExceptionStatusRequest(**kwargs)
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {"function_code": resp.function_code, "status": resp.status}
        return ExtendedRequestSupport._process_exception(resp)

    def get_com_event_counter(self, **kwargs):
        """Read status word and an event count.

        From the remote device"s communication event counter.

        :param kwargs:
        :return:
        """
        request = GetCommEventCounterRequest(**kwargs)
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {
                "function_code": resp.function_code,
                "status": resp.status,
                "count": resp.count,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def get_com_event_log(self, **kwargs):
        """Read status word.

        Event count, message count, and a field of event
        bytes from the remote device.

        :param kwargs:
        :return:
        """
        request = GetCommEventLogRequest(**kwargs)
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {
                "function_code": resp.function_code,
                "status": resp.status,
                "message count": resp.message_count,
                "event count": resp.event_count,
                "events": resp.events,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def _execute_diagnostic_request(self, request):
        """Execute diagnostic request."""
        resp = self.execute(request)  # pylint: disable=no-member
        if not resp.isError():
            return {
                "function code": resp.function_code,
                "sub function code": resp.sub_function_code,
                "message": resp.message,
            }
        return ExtendedRequestSupport._process_exception(resp)

    def return_query_data(self, message=0, **kwargs):
        """Loop back data sent in response.

        :param message: Message to be looped back
        :param kwargs:
        :return:
        """
        request = ReturnQueryDataRequest(message, **kwargs)
        return self._execute_diagnostic_request(request)

    def restart_comm_option(self, toggle=False, **kwargs):
        """Initialize and restart remote devices.

        Serial interface and clear all of its communications event counters.

        :param toggle: Toggle Status [ON(0xff00)/OFF(0x0000]
        :param kwargs:
        :return:
        """
        request = RestartCommunicationsOptionRequest(toggle, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_diagnostic_register(self, data=0, **kwargs):
        """Read 16-bit diagnostic register.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnDiagnosticRegisterRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def change_ascii_input_delimiter(self, data=0, **kwargs):
        """Change message delimiter for future requests.

        :param data: New delimiter character
        :param kwargs:
        :return:
        """
        request = ChangeAsciiInputDelimiterRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def force_listen_only_mode(self, data=0, **kwargs):
        """Force addressed remote device to its Listen Only Mode.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ForceListenOnlyModeRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def clear_counters(self, data=0, **kwargs):
        """Clear all counters and diag registers.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ClearCountersRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_bus_message_count(self, data=0, **kwargs):
        """Return count of message detected on bus by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnBusMessageCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_bus_com_error_count(self, data=0, **kwargs):
        """Return count of CRC errors received by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnBusCommunicationErrorCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_bus_exception_error_count(self, data=0, **kwargs):
        """Return count of Modbus exceptions returned by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnBusExceptionErrorCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_slave_message_count(self, data=0, **kwargs):
        """Return count of messages addressed to remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnSlaveMessageCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_slave_no_response_count(self, data=0, **kwargs):
        """Return count of No responses by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnSlaveNoResponseCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_slave_no_ack_count(self, data=0, **kwargs):
        """Return count of NO ACK exceptions sent by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnSlaveNAKCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_slave_busy_count(self, data=0, **kwargs):
        """Return count of server busy exceptions sent by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnSlaveBusyCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_slave_bus_char_overrun_count(self, data=0, **kwargs):
        """Return count of messages not handled.

        By remote slave due to character overrun condition.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnSlaveBusCharacterOverrunCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def return_iop_overrun_count(self, data=0, **kwargs):
        """Return count of iop overrun errors by remote slave.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ReturnIopOverrunCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def clear_overrun_count(self, data=0, **kwargs):
        """Clear over run counter.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = ClearOverrunCountRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)

    def get_clear_modbus_plus(self, data=0, **kwargs):
        """Get/clear stats of remote modbus plus device.

        :param data: Data field (0x0000)
        :param kwargs:
        :return:
        """
        request = GetClearModbusPlusRequest(data, **kwargs)
        return self._execute_diagnostic_request(request)


class ModbusSerialClient(ExtendedRequestSupport, _ModbusSerialClient):
    """Modbus serial client."""

    def __init__(self, framer, **kwargs):
        """Initialize."""
        super().__init__(framer=framer, **kwargs)

    def get_port(self):
        """Get serial Port.

        :return: Current Serial port
        """
        return self.params.port

    def set_port(self, value):
        """Set serial Port setter.

        :param value: New port
        """
        self.params.port = value
        if self.is_socket_open():
            self.close()

    def get_stopbits(self):
        """Get number of stop bits.

        :return: Current Stop bits
        """
        return self.params.stopbits

    def set_stopbits(self, value):
        """Set stop bit.

        :param value: Possible values (1, 1.5, 2)
        """
        self.params.stopbits = float(value)
        if self.is_socket_open():
            self.close()

    def get_bytesize(self):
        """Get number of data bits.

        :return: Current bytesize
        """
        return self.params.bytesize

    def set_bytesize(self, value):
        """Set Byte size.

        :param value: Possible values (5, 6, 7, 8)

        """
        self.params.bytesize = int(value)
        if self.is_socket_open():
            self.close()

    def get_parity(self):
        """Enable Parity Checking.

        :return: Current parity setting
        """
        return self.params.parity

    def set_parity(self, value):
        """Set parity Setter.

        :param value: Possible values ("N", "E", "O", "M", "S")
        """
        self.params.parity = value
        if self.is_socket_open():
            self.close()

    def get_baudrate(self):
        """Get serial Port baudrate.

        :return: Current baudrate
        """
        return self.params.baudrate

    def set_baudrate(self, value):
        """Set baudrate setter.

        :param value: <supported baudrate>
        """
        self.params.baudrate = int(value)
        if self.is_socket_open():
            self.close()

    def get_timeout(self):
        """Get serial Port Read timeout.

        :return: Current read imeout.
        """
        return self.params.timeout

    def set_timeout(self, value):
        """Read timeout setter.

        :param value: Read Timeout in seconds
        """
        self.params.timeout = float(value)
        if self.is_socket_open():
            self.close()

    def get_serial_settings(self):
        """Get Current Serial port settings.

        :return: Current Serial settings as dict.
        """
        return {
            "baudrate": self.params.baudrate,
            "port": self.params.port,
            "parity": self.params.parity,
            "stopbits": self.params.stopbits,
            "bytesize": self.params.bytesize,
            "read timeout": self.params.timeout,
            "t1.5": self.inter_char_timeout,
            "t3.5": self.silent_interval,
        }


class ModbusTcpClient(ExtendedRequestSupport, _ModbusTcpClient):
    """TCP client."""

    def __init__(self, **kwargs):
        """Initialize."""
        super().__init__(**kwargs)



"""Test register read messages."""
from test.conftest import FakeList, MockContext
import unittest

from pymodbus.pdu import ModbusExceptions
from pymodbus.register_read_message import (
    ReadHoldingRegistersRequest,
    ReadHoldingRegistersResponse,
    ReadInputRegistersRequest,
    ReadInputRegistersResponse,
    ReadRegistersRequestBase,
    ReadRegistersResponseBase,
    ReadWriteMultipleRegistersRequest,
    ReadWriteMultipleRegistersResponse,
)


TEST_MESSAGE = b"\x06\x00\x0a\x00\x0b\x00\x0c"

# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class ReadRegisterMessagesTest(unittest.TestCase):
    """Register Message Test Fixture.

    This fixture tests the functionality of all the
    register based request/response messages:

    * Read/Write Input Registers
    * Read Holding Registers
    """

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""
        arguments = {
            "read_address": 1,
            "read_count": 5,
            "write_address": 1,
            "write_registers": [0x00] * 5,
        }
        self.value = 0xABCD
        self.values = [0xA, 0xB, 0xC]
        self.request_read = {
            ReadRegistersRequestBase(1, 5): b"\x00\x01\x00\x05",
            ReadHoldingRegistersRequest(1, 5): b"\x00\x01\x00\x05",
            ReadInputRegistersRequest(1, 5): b"\x00\x01\x00\x05",
            ReadWriteMultipleRegistersRequest(
                **arguments
            ): b"\x00\x01\x00\x05\x00\x01\x00"
            b"\x05\x0a\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00",
        }
        self.response_read = {
            ReadRegistersResponseBase(self.values): TEST_MESSAGE,
            ReadHoldingRegistersResponse(self.values): TEST_MESSAGE,
            ReadInputRegistersResponse(self.values): TEST_MESSAGE,
            ReadWriteMultipleRegistersResponse(self.values): TEST_MESSAGE,
        }

    def tearDown(self):
        """Clean up the test environment."""
        del self.request_read
        del self.response_read

    def test_read_register_response_base(self):
        """Test read register response."""
        response = ReadRegistersResponseBase(list(range(10)))
        for index in range(10):
            self.assertEqual(response.getRegister(index), index)

    def test_register_read_requests(self):
        """Test register read requests."""
        for request, response in iter(self.request_read.items()):
            self.assertEqual(request.encode(), response)

    def test_register_read_responses(self):
        """Test register read response."""
        for request, response in iter(self.response_read.items()):
            self.assertEqual(request.encode(), response)

    def test_register_read_response_decode(self):
        """Test register read response."""
        registers = [
            [0x0A, 0x0B, 0x0C],
            [0x0A, 0x0B, 0x0C],
            [0x0A, 0x0B, 0x0C],
            [0x0A, 0x0B, 0x0C, 0x0A, 0x0B, 0x0C],
        ]
        values = sorted(
            self.response_read.items(),
            key=lambda x: str(x),  # pylint: disable=unnecessary-lambda
        )
        for packet, register in zip(values, registers):
            request, response = packet
            request.decode(response)
            self.assertEqual(request.registers, register)

    def test_register_read_requests_count_errors(self):
        """This tests that the register request messages.

        will break on counts that are out of range
        """
        mock = FakeList(0x800)
        requests = [
            ReadHoldingRegistersRequest(1, 0x800),
            ReadInputRegistersRequest(1, 0x800),
            ReadWriteMultipleRegistersRequest(
                read_address=1, read_count=0x800, write_address=1, write_registers=5
            ),
            ReadWriteMultipleRegistersRequest(
                read_address=1, read_count=5, write_address=1, write_registers=mock
            ),
        ]
        for request in requests:
            result = request.execute(None)
            self.assertEqual(ModbusExceptions.IllegalValue, result.exception_code)

    def test_register_read_requests_validate_errors(self):
        """This tests that the register request messages.

        will break on counts that are out of range
        """
        context = MockContext()
        requests = [
            ReadHoldingRegistersRequest(-1, 5),
            ReadInputRegistersRequest(-1, 5),
            # ReadWriteMultipleRegistersRequest(-1,5,1,5),
            # ReadWriteMultipleRegistersRequest(1,5,-1,5),
        ]
        for request in requests:
            result = request.execute(context)
            self.assertEqual(ModbusExceptions.IllegalAddress, result.exception_code)

    def test_register_read_requests_execute(self):
        """This tests that the register request messages.

        will break on counts that are out of range
        """
        context = MockContext(True)
        requests = [
            ReadHoldingRegistersRequest(-1, 5),
            ReadInputRegistersRequest(-1, 5),
        ]
        for request in requests:
            response = request.execute(context)
            self.assertEqual(request.function_code, response.function_code)

    def test_read_write_multiple_registers_request(self):
        """Test read/write multiple registers."""
        context = MockContext(True)
        request = ReadWriteMultipleRegistersRequest(
            read_address=1, read_count=10, write_address=1, write_registers=[0x00]
        )
        response = request.execute(context)
        self.assertEqual(request.function_code, response.function_code)

    def test_read_write_multiple_registers_validate(self):
        """Test read/write multiple registers."""
        context = MockContext()
        context.validate = lambda f, a, c: a == 1
        request = ReadWriteMultipleRegistersRequest(
            read_address=1, read_count=10, write_address=2, write_registers=[0x00]
        )
        response = request.execute(context)
        self.assertEqual(response.exception_code, ModbusExceptions.IllegalAddress)

        context.validate = lambda f, a, c: a == 2
        response = request.execute(context)
        self.assertEqual(response.exception_code, ModbusExceptions.IllegalAddress)

        request.write_byte_count = 0x100
        response = request.execute(context)
        self.assertEqual(response.exception_code, ModbusExceptions.IllegalValue)

    def test_read_write_multiple_registers_request_decode(self):
        """Test read/write multiple registers."""
        request, response = next(
            (k, v)
            for k, v in self.request_read.items()
            if getattr(k, "function_code", 0) == 23
        )
        request.decode(response)
        self.assertEqual(request.read_address, 0x01)
        self.assertEqual(request.write_address, 0x01)
        self.assertEqual(request.read_count, 0x05)
        self.assertEqual(request.write_count, 0x05)
        self.assertEqual(request.write_byte_count, 0x0A)
        self.assertEqual(request.write_registers, [0x00] * 5)

    def test_serializing_to_string(self):
        """Test serializing to string."""
        for request in iter(self.request_read.keys()):
            self.assertTrue(str(request) is not None)
        for request in iter(self.response_read.keys()):
            self.assertTrue(str(request) is not None)



"""Test factory."""
import unittest

from pymodbus.exceptions import ModbusException, MessageRegisterException
from pymodbus.factory import ClientDecoder, ServerDecoder
from pymodbus.pdu import ModbusRequest, ModbusResponse


def _raise_exception(_):
    """Raise exception."""
    raise ModbusException("something")


class SimpleFactoryTest(unittest.TestCase):
    """Unittest for the pymod.exceptions module."""

    def setUp(self):
        """Initialize the test environment"""
        self.client = ClientDecoder()
        self.server = ServerDecoder()
        self.request = (
            (0x01, b"\x01\x00\x01\x00\x01"),  # read coils
            (0x02, b"\x02\x00\x01\x00\x01"),  # read discrete inputs
            (0x03, b"\x03\x00\x01\x00\x01"),  # read holding registers
            (0x04, b"\x04\x00\x01\x00\x01"),  # read input registers
            (0x05, b"\x05\x00\x01\x00\x01"),  # write single coil
            (0x06, b"\x06\x00\x01\x00\x01"),  # write single register
            (0x07, b"\x07"),  # read exception status
            (0x08, b"\x08\x00\x00\x00\x00"),  # read diagnostic
            (0x0B, b"\x0b"),  # get comm event counters
            (0x0C, b"\x0c"),  # get comm event log
            (0x0F, b"\x0f\x00\x01\x00\x08\x01\x00\xff"),  # write multiple coils
            (0x10, b"\x10\x00\x01\x00\x02\x04\0xff\xff"),  # write multiple registers
            (0x11, b"\x11"),  # report slave id
            (
                0x14,
                b"\x14\x0e\x06\x00\x04\x00\x01\x00\x02\x06\x00\x03\x00\x09\x00\x02",
            ),  # read file record
            (
                0x15,
                b"\x15\x0d\x06\x00\x04\x00\x07\x00\x03\x06\xaf\x04\xbe\x10\x0d",
            ),  # write file record
            (0x16, b"\x16\x00\x01\x00\xff\xff\x00"),  # mask write register
            (
                0x17,
                b"\x17\x00\x01\x00\x01\x00\x01\x00\x01\x02\x12\x34",
            ),  # r/w multiple regs
            (0x18, b"\x18\x00\x01"),  # read fifo queue
            (0x2B, b"\x2b\x0e\x01\x00"),  # read device identification
        )

        self.response = (
            (0x01, b"\x01\x01\x01"),  # read coils
            (0x02, b"\x02\x01\x01"),  # read discrete inputs
            (0x03, b"\x03\x02\x01\x01"),  # read holding registers
            (0x04, b"\x04\x02\x01\x01"),  # read input registers
            (0x05, b"\x05\x00\x01\x00\x01"),  # write single coil
            (0x06, b"\x06\x00\x01\x00\x01"),  # write single register
            (0x07, b"\x07\x00"),  # read exception status
            (0x08, b"\x08\x00\x00\x00\x00"),  # read diagnostic
            (0x0B, b"\x0b\x00\x00\x00\x00"),  # get comm event counters
            (0x0C, b"\x0c\x08\x00\x00\x01\x08\x01\x21\x20\x00"),  # get comm event log
            (0x0F, b"\x0f\x00\x01\x00\x08"),  # write multiple coils
            (0x10, b"\x10\x00\x01\x00\x02"),  # write multiple registers
            (0x11, b"\x11\x03\x05\x01\x54"),  # report slave id (device specific)
            (
                0x14,
                b"\x14\x0c\x05\x06\x0d\xfe\x00\x20\x05\x06\x33\xcd\x00\x40",
            ),  # read file record
            (
                0x15,
                b"\x15\x0d\x06\x00\x04\x00\x07\x00\x03\x06\xaf\x04\xbe\x10\x0d",
            ),  # write file record
            (0x16, b"\x16\x00\x01\x00\xff\xff\x00"),  # mask write register
            (0x17, b"\x17\x02\x12\x34"),  # read/write multiple registers
            (0x18, b"\x18\x00\x01\x00\x01\x00\x00"),  # read fifo queue
            (
                0x2B,
                b"\x2b\x0e\x01\x01\x00\x00\x01\x00\x01\x77",
            ),  # read device identification
        )

        self.exception = (
            (0x81, b"\x81\x01\xd0\x50"),  # illegal function exception
            (0x82, b"\x82\x02\x90\xa1"),  # illegal data address exception
            (0x83, b"\x83\x03\x50\xf1"),  # illegal data value exception
            (0x84, b"\x84\x04\x13\x03"),  # skave device failure exception
            (0x85, b"\x85\x05\xd3\x53"),  # acknowledge exception
            (0x86, b"\x86\x06\x93\xa2"),  # slave device busy exception
            (0x87, b"\x87\x08\x53\xf2"),  # memory parity exception
            (0x88, b"\x88\x0a\x16\x06"),  # gateway path unavailable exception
            (0x89, b"\x89\x0b\xd6\x56"),  # gateway target failed exception
        )

        self.bad = (
            (0x80, b"\x80\x00\x00\x00"),  # Unknown Function
            (0x81, b"\x81\x00\x00\x00"),  # error message
        )

    def tearDown(self):
        """Clean up the test environment"""
        del self.bad
        del self.request
        del self.response

    def test_exception_lookup(self):
        """Test that we can look up exception messages"""
        for func, _ in self.exception:
            response = self.client.lookupPduClass(func)
            self.assertNotEqual(response, None)

        for func, _ in self.exception:
            response = self.server.lookupPduClass(func)
            self.assertNotEqual(response, None)

    def test_response_lookup(self):
        """Test a working response factory lookup"""
        for func, _ in self.response:
            response = self.client.lookupPduClass(func)
            self.assertNotEqual(response, None)

    def test_request_ookup(self):
        """Test a working request factory lookup"""
        for func, _ in self.request:
            request = self.client.lookupPduClass(func)
            self.assertNotEqual(request, None)

    def test_response_working(self):
        """Test a working response factory decoders"""
        for func, msg in self.response:
            self.client.decode(msg)

    def test_response_errors(self):
        """Test a response factory decoder exceptions"""
        self.assertRaises(
            ModbusException,
            self.client._helper,  # pylint: disable=protected-access
            self.bad[0][1],
        )
        self.assertEqual(
            self.client.decode(self.bad[1][1]).function_code,
            self.bad[1][0],
            "Failed to decode error PDU",
        )

    def test_requests_working(self):
        """Test a working request factory decoders"""
        for func, msg in self.request:
            self.server.decode(msg)

    def test_client_factory_fails(self):
        """Tests that a client factory will fail to decode a bad message"""
        self.client._helper = _raise_exception  # pylint: disable=protected-access
        actual = self.client.decode(None)
        self.assertEqual(actual, None)

    def test_server_factory_fails(self):
        """Tests that a server factory will fail to decode a bad message"""
        self.server._helper = _raise_exception  # pylint: disable=protected-access
        actual = self.server.decode(None)
        self.assertEqual(actual, None)

    def test_server_register_custom_request(self):
        """Test server register custom request."""

        class CustomRequest(ModbusRequest):
            """Custom request."""

            function_code = 0xFF

        class NoCustomRequest():  # pylint: disable=too-few-public-methods
            """Custom request."""

            function_code = 0xFF

        self.server.register(CustomRequest)
        self.assertTrue(self.client.lookupPduClass(CustomRequest.function_code))
        CustomRequest.sub_function_code = 0xFF
        self.server.register(CustomRequest)
        self.assertTrue(self.server.lookupPduClass(CustomRequest.function_code))
        try:
            func_raised = False
            self.server.register(NoCustomRequest)
        except MessageRegisterException:
            func_raised = True
        self.assertTrue(func_raised)

    def test_client_register_custom_response(self):
        """Test client register custom response."""

        class CustomResponse(ModbusResponse):
            """Custom response."""

            function_code = 0xFF

        class NoCustomResponse():  # pylint: disable=too-few-public-methods
            """Custom request."""

            function_code = 0xFF

        self.client.register(CustomResponse)
        self.assertTrue(self.client.lookupPduClass(CustomResponse.function_code))
        CustomResponse.sub_function_code = 0xFF
        self.client.register(CustomResponse)
        self.assertTrue(self.client.lookupPduClass(CustomResponse.function_code))
        try:
            func_raised = False
            self.client.register(NoCustomResponse)
        except MessageRegisterException:
            func_raised = True
        self.assertTrue(func_raised)

    # ---------------------------------------------------------------------------#
    #  I don"t actually know what is supposed to be returned here, I assume that
    #  since the high bit is set, it will simply echo the resulting message
    # ---------------------------------------------------------------------------#

    def test_request_errors(self):
        """Test a request factory decoder exceptions"""
        for func, msg in self.bad:
            result = self.server.decode(msg)
            self.assertEqual(result.ErrorCode, 1, "Failed to decode invalid requests")
            self.assertEqual(
                result.execute(None).function_code,
                func,
                "Failed to create correct response message",
            )



# pylint: disable=missing-type-doc,missing-param-doc,differing-param-doc,missing-raises-doc,missing-any-param-doc
"""Modbus BCD Payload Builder.

This is an example of building a custom payload builder
that can be used in the pymodbus library. Below is a
simple binary coded decimal builder and decoder.
"""
from struct import pack

from pymodbus.constants import Endian
from pymodbus.exceptions import ParameterException
from pymodbus.interfaces import IPayloadBuilder
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.utilities import pack_bitstring, unpack_bitstring


def convert_to_bcd(decimal):
    """Convert a decimal value to a bcd value

    :param value: The decimal value to to pack into bcd
    :returns: The number in bcd form
    """
    place, bcd = 0, 0
    while decimal > 0:
        nibble = decimal % 10
        bcd += nibble << place
        decimal /= 10
        place += 4
    return bcd


def convert_from_bcd(bcd):
    """Convert a bcd value to a decimal value

    :param value: The value to unpack from bcd
    :returns: The number in decimal form
    """
    place, decimal = 1, 0
    while bcd > 0:
        nibble = bcd & 0xF
        decimal += nibble * place
        bcd >>= 4
        place *= 10
    return decimal


def count_bcd_digits(bcd):
    """Count the number of digits in a bcd value

    :param bcd: The bcd number to count the digits of
    :returns: The number of digits in the bcd string
    """
    count = 0
    while bcd > 0:
        count += 1
        bcd >>= 4
    return count


class BcdPayloadBuilder(IPayloadBuilder):
    """A utility that helps build binary coded decimal payload messages

    to be written with the various modbus messages.
    example::

        builder = BcdPayloadBuilder()
        builder.add_number(1)
        builder.add_number(int(2.234 * 1000))
        payload = builder.build()
    """

    def __init__(self, payload=None, endian=Endian.Little):
        """Initialize a new instance of the payload builder

        :param payload: Raw payload data to initialize with
        :param endian: The endianness of the payload
        """
        self._endian = endian
        self._payload = payload or []

    def __str__(self):
        """Return the payload buffer as a string

        :returns: The payload buffer as a string
        """
        return "".join(self._payload)

    def reset(self):
        """Reset the payload buffer"""
        self._payload = []

    def build(self):
        """Return the payload buffer as a list

        This list is two bytes per element and can
        thus be treated as a list of registers.

        :returns: The payload buffer as a list
        """
        string = str(self)
        length = len(string)
        string = string + ("\x00" * (length % 2))
        return [string[i : i + 2] for i in range(0, length, 2)]

    def add_bits(self, values):
        """Add a collection of bits to be encoded

        If these are less than a multiple of eight,
        they will be left padded with 0 bits to make
        it so.

        :param value: The value to add to the buffer
        """
        value = pack_bitstring(values)
        self._payload.append(value)

    def add_number(self, value, size=None):
        """Add any 8bit numeric type to the buffer

        :param value: The value to add to the buffer
        """
        encoded = []
        value = convert_to_bcd(value)
        size = size or count_bcd_digits(value)
        while size > 0:
            nibble = value & 0xF
            encoded.append(pack("B", nibble))
            value >>= 4
            size -= 1
        self._payload.extend(encoded)

    def add_string(self, value):
        """Add a string to the buffer

        :param value: The value to add to the buffer
        """
        self._payload.append(value)


class BcdPayloadDecoder:
    """A utility that helps decode binary coded decimal payload messages from a modbus response message.

    What follows is a simple example::

        decoder = BcdPayloadDecoder(payload)
        first   = decoder.decode_int(2)
        second  = decoder.decode_int(5) / 100
    """

    def __init__(self, payload):
        """Initialize a new payload decoder

        :param payload: The payload to decode with
        """
        self._payload = payload
        self._pointer = 0x00

    @staticmethod
    def fromRegisters(registers, endian=Endian.Little):  # pylint: disable=invalid-name
        """Initialize a payload decoder

        with the result of reading a collection of registers from a modbus device.

        The registers are treated as a list of 2 byte values.
        We have to do this because of how the data has already
        been decoded by the rest of the library.

        :param registers: The register results to initialize with
        :param endian: The endianness of the payload
        :returns: An initialized PayloadDecoder
        """
        if isinstance(registers, list):  # repack into flat binary
            payload = "".join(pack(">H", x) for x in registers)
            return BinaryPayloadDecoder(payload, endian)
        raise ParameterException("Invalid collection of registers supplied")

    @staticmethod
    def fromCoils(coils, endian=Endian.Little):  # pylint: disable=invalid-name
        """Initialize a payload decoder.

        with the result of reading a collection of coils from a modbus device.

        The coils are treated as a list of bit(boolean) values.

        :param coils: The coil results to initialize with
        :param endian: The endianness of the payload
        :returns: An initialized PayloadDecoder
        """
        if isinstance(coils, list):
            payload = pack_bitstring(coils)
            return BinaryPayloadDecoder(payload, endian)
        raise ParameterException("Invalid collection of coils supplied")

    def reset(self):
        """Reset the decoder pointer back to the start"""
        self._pointer = 0x00

    def decode_int(self, size=1):
        """Decode a int or long from the buffer"""
        self._pointer += size
        handle = self._payload[self._pointer - size : self._pointer]
        return convert_from_bcd(handle)

    def decode_bits(self):
        """Decode a byte worth of bits from the buffer"""
        self._pointer += 1
        handle = self._payload[self._pointer - 1 : self._pointer]
        return unpack_bitstring(handle)

    def decode_string(self, size=1):
        """Decode a string from the buffer

        :param size: The size of the string to decode
        """
        self._pointer += size
        return self._payload[self._pointer - size : self._pointer]


# --------------------------------------------------------------------------- #
# Exported Identifiers
# --------------------------------------------------------------------------- #

__all__ = ["BcdPayloadBuilder", "BcdPayloadDecoder"]



"""Test exceptions."""
import unittest

from pymodbus.exceptions import (
    ConnectionException,
    ModbusException,
    ModbusIOException,
    NotImplementedException,
    ParameterException,
)


class SimpleExceptionsTest(unittest.TestCase):
    """Unittest for the pymodbus.exceptions module."""

    def setUp(self):
        """Initialize the test environment"""
        self.exceptions = [
            ModbusException("bad base"),
            ModbusIOException("bad register"),
            ParameterException("bad parameter"),
            NotImplementedException("bad function"),
            ConnectionException("bad connection"),
        ]

    def tearDown(self):
        """Clean up the test environment"""

    def test_exceptions(self):
        """Test all module exceptions"""
        for exc in self.exceptions:
            try:
                raise exc
            except ModbusException as exc:
                self.assertTrue("Modbus Error:" in str(exc))
                return
            self.fail("Excepted a ModbusExceptions")



#!/usr/bin/env python3
"""Pymodbus Logging Examples."""
import logging
import logging.handlers as Handlers


if __name__ == "__main__":
    # ----------------------------------------------------------------------- #
    # This will simply send everything logged to console
    # ----------------------------------------------------------------------- #
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)

    # ----------------------------------------------------------------------- #
    # This will send the error messages in the specified namespace to a file.
    # The available namespaces in pymodbus are as follows:
    # ----------------------------------------------------------------------- #
    # * pymodbus.*          - The root namespace
    # * pymodbus.server.*   - all logging messages involving the modbus server
    # * pymodbus.client.*   - all logging messages involving the client
    # * pymodbus.protocol.* - all logging messages inside the protocol layer
    # ----------------------------------------------------------------------- #
    log = logging.getLogger("pymodbus.server")
    log.setLevel(logging.ERROR)

    # ----------------------------------------------------------------------- #
    # This will send the error messages to the specified handlers:
    # * docs.python.org/library/logging.html
    # ----------------------------------------------------------------------- #
    log = logging.getLogger("pymodbus")
    log.setLevel(logging.ERROR)
    handlers = [
        Handlers.RotatingFileHandler("logfile", maxBytes=1024 * 1024),
        Handlers.SMTPHandler(
            "mx.host.com", "pymodbus@host.com", ["support@host.com"], "Pymodbus"
        ),
        Handlers.SysLogHandler(facility="daemon"),
        Handlers.DatagramHandler("localhost", 12345),
    ]
    [log.addHandler(h) for h in handlers]  # pylint: disable=expression-not-assigned



"""Test events."""
import unittest

from pymodbus.events import (
    CommunicationRestartEvent,
    EnteredListenModeEvent,
    ModbusEvent,
    RemoteReceiveEvent,
    RemoteSendEvent,
)
from pymodbus.exceptions import NotImplementedException, ParameterException


class ModbusEventsTest(unittest.TestCase):
    """Unittest for the pymodbus.device module."""

    def setUp(self):
        """Set up the test environment"""

    def tearDown(self):
        """Clean up the test environment"""

    def test_modbus_event_base_class(self):
        """Test modbus event base class."""
        event = ModbusEvent()
        self.assertRaises(NotImplementedException, event.encode)
        self.assertRaises(NotImplementedException, lambda: event.decode(None))

    def test_remote_receive_event(self):
        """Test remove receive event."""
        event = RemoteReceiveEvent()
        event.decode(b"\x70")
        self.assertTrue(event.overrun)
        self.assertTrue(event.listen)
        self.assertTrue(event.broadcast)

    def test_remote_sent_event(self):
        """Test remote sent event."""
        event = RemoteSendEvent()
        result = event.encode()
        self.assertEqual(result, b"\x40")
        event.decode(b"\x7f")
        self.assertTrue(event.read)
        self.assertTrue(event.slave_abort)
        self.assertTrue(event.slave_busy)
        self.assertTrue(event.slave_nak)
        self.assertTrue(event.write_timeout)
        self.assertTrue(event.listen)

    def test_remote_sent_event_encode(self):
        """Test remote sent event encode."""
        arguments = {
            "read": True,
            "slave_abort": True,
            "slave_busy": True,
            "slave_nak": True,
            "write_timeout": True,
            "listen": True,
        }
        event = RemoteSendEvent(**arguments)
        result = event.encode()
        self.assertEqual(result, b"\x7f")

    def test_entered_listen_mode_event(self):
        """Test entered listen mode event."""
        event = EnteredListenModeEvent()
        result = event.encode()
        self.assertEqual(result, b"\x04")
        event.decode(b"\x04")
        self.assertEqual(event.value, 0x04)
        self.assertRaises(ParameterException, lambda: event.decode(b"\x00"))

    def test_communication_restart_event(self):
        """Test communication restart event."""
        event = CommunicationRestartEvent()
        result = event.encode()
        self.assertEqual(result, b"\x00")
        event.decode(b"\x00")
        self.assertEqual(event.value, 0x00)
        self.assertRaises(ParameterException, lambda: event.decode(b"\x04"))



"""Helper Module for REPL actions."""
from collections import OrderedDict
import inspect

# pylint: disable=missing-type-doc
import json

from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML, PygmentsTokens
import pygments
from pygments.lexers.data import JsonLexer

from pymodbus.payload import BinaryPayloadDecoder, Endian


predicate = inspect.isfunction
argspec = inspect.signature


FORMATTERS = {
    "int8": "decode_8bit_int",
    "int16": "decode_16bit_int",
    "int32": "decode_32bit_int",
    "int64": "decode_64bit_int",
    "uint8": "decode_8bit_uint",
    "uint16": "decode_16bit_uint",
    "uint32": "decode_32bit_uint",
    "uint64": "decode_64bit_int",
    "float16": "decode_16bit_float",
    "float32": "decode_32bit_float",
    "float64": "decode_64bit_float",
}


DEFAULT_KWARGS = {"unit": "Slave address"}

OTHER_COMMANDS = {
    "result.raw": "Show RAW Result",
    "result.decode": "Decode register response to known formats",
}
EXCLUDE = ["execute", "recv", "send", "trace", "set_debug"]
CLIENT_METHODS = [
    "connect",
    "close",
    "idle_time",
    "is_socket_open",
    "get_port",
    "set_port",
    "get_stopbits",
    "set_stopbits",
    "get_bytesize",
    "set_bytesize",
    "get_parity",
    "set_parity",
    "get_baudrate",
    "set_baudrate",
    "get_timeout",
    "set_timeout",
    "get_serial_settings",
]
CLIENT_ATTRIBUTES = []


class Command:
    """Class representing Commands to be consumed by Completer."""

    def __init__(self, name, signature, doc, unit=False):
        """Initialize.

        :param name: Name of the command
        :param signature: inspect object
        :param doc: Doc string for the command
        :param unit: Use unit as additional argument in the command .
        """
        self.name = name
        self.doc = doc.split("\n") if doc else " ".join(name.split("_"))
        self.help_text = self._create_help()
        self.param_help = self._create_arg_help()
        if signature:
            self._params = signature.parameters
            self.args = self.create_completion()
        else:
            self._params = ""

        if self.name.startswith("client.") and unit:
            self.args.update(**DEFAULT_KWARGS)

    def _create_help(self):
        """Create help."""
        doc = filter(lambda d: d, self.doc)
        cmd_help = list(
            filter(
                lambda x: not x.startswith(":param") and not x.startswith(":return"),
                doc,
            )
        )
        return " ".join(cmd_help).strip()

    def _create_arg_help(self):
        """Create arg help."""
        param_dict = {}
        params = list(filter(lambda d: d.strip().startswith(":param"), self.doc))
        for param in params:
            param, param_help = param.split(":param")[1].strip().split(":")
            param_dict[param] = param_help
        return param_dict

    def create_completion(self):
        """Create command completion meta data.

        :return:
        """
        words = {}

        def _create(entry, default):
            if entry not in ["self", "kwargs"]:
                if isinstance(default, (int, str)):
                    entry += f"={default}"
                return entry
            return None

        for arg in self._params.values():
            if entry := _create(arg.name, arg.default):
                entry, meta = self.get_meta(entry)
                words[entry] = meta

        return words

    def get_completion(self):
        """Get a list of completions.

        :return:
        """
        return self.args.keys()

    def get_meta(self, cmd):
        """Get Meta info of a given command.

        :param cmd: Name of command.
        :return: Dict containing meta info.
        """
        cmd = cmd.strip()
        cmd = cmd.split("=")[0].strip()
        return cmd, self.param_help.get(cmd, "")

    def __str__(self):
        """Return string representation."""
        if self.doc:
            return "Command {:>50}{:<20}".format(  # pylint: disable=consider-using-f-string
                self.name, self.doc
            )
        return f"Command {self.name}"


def _get_requests(members):
    """Get requests."""
    commands = list(
        filter(
            lambda x: (
                x[0] not in EXCLUDE and x[0] not in CLIENT_METHODS and callable(x[1])
            ),
            members,
        )
    )
    commands = {
        f"client.{c[0]}": Command(
            f"client.{c[0]}", argspec(c[1]), inspect.getdoc(c[1]), unit=True
        )
        for c in commands
        if not c[0].startswith("_")
    }
    return commands


def _get_client_methods(members):
    """Get client methods."""
    commands = list(
        filter(lambda x: (x[0] not in EXCLUDE and x[0] in CLIENT_METHODS), members)
    )
    commands = {
        "client.{c[0]}": Command(
            "client.{c[0]}", argspec(c[1]), inspect.getdoc(c[1]), unit=False
        )
        for c in commands
        if not c[0].startswith("_")
    }
    return commands


def _get_client_properties(members):
    """Get client properties."""
    global CLIENT_ATTRIBUTES  # pylint: disable=global-variable-not-assigned
    commands = list(filter(lambda x: not callable(x[1]), members))
    commands = {
        f"client.{c[0]}": Command(f"client.{c[0]}", None, "Read Only!", unit=False)
        for c in commands
        if (not c[0].startswith("_") and isinstance(c[1], (str, int, float)))
    }
    CLIENT_ATTRIBUTES.extend(list(commands.keys()))
    return commands


def get_commands(client):
    """Retrieve all required methods and attributes.

    Of a client object and convert it to commands.

    :param client: Modbus Client object.
    :return:
    """
    commands = {}
    members = inspect.getmembers(client)
    requests = _get_requests(members)
    client_methods = _get_client_methods(members)
    client_attr = _get_client_properties(members)

    result_commands = inspect.getmembers(Result, predicate=predicate)
    result_commands = {
        f"result.{c[0]}": Command(f"result.{c[0]}", argspec(c[1]), inspect.getdoc(c[1]))
        for c in result_commands
        if (not c[0].startswith("_") and c[0] != "print_result")
    }
    commands.update(requests)
    commands.update(client_methods)
    commands.update(client_attr)
    commands.update(result_commands)
    return commands


class Result:
    """Represent result command."""

    function_code = None
    data = None

    def __init__(self, result):
        """Initialize.

        :param result: Response of a modbus command.
        """
        if isinstance(result, dict):  # Modbus response
            self.function_code = result.pop("function_code", None)
            self.data = dict(result)
        else:
            self.data = result

    def decode(self, formatters, byte_order="big", word_order="big"):
        """Decode the register response to known formatters.

        :param formatters: int8/16/32/64, uint8/16/32/64, float32/64
        :param byte_order: little/big
        :param word_order: little/big
        """
        # Read Holding Registers (3)
        # Read Input Registers (4)
        # Read Write Registers (23)
        if not isinstance(formatters, (list, tuple)):
            formatters = [formatters]

        if self.function_code not in [3, 4, 23]:
            print_formatted_text(HTML("<red>Decoder works only for registers!!</red>"))
            return
        byte_order = (
            Endian.Little if byte_order.strip().lower() == "little" else Endian.Big
        )
        word_order = (
            Endian.Little if word_order.strip().lower() == "little" else Endian.Big
        )
        decoder = BinaryPayloadDecoder.fromRegisters(
            self.data.get("registers"), byteorder=byte_order, wordorder=word_order
        )
        for formatter in formatters:
            if not (formatter := FORMATTERS.get(formatter)):
                print_formatted_text(
                    HTML(f"<red>Invalid Formatter - {formatter}!!</red>")
                )
                return
            decoded = getattr(decoder, formatter)()
            self.print_result(decoded)

    def raw(self):
        """Return raw result dict."""
        self.print_result()

    def _process_dict(self, use_dict):
        """Process dict."""
        new_dict = OrderedDict()
        for k, v_item in use_dict.items():
            if isinstance(v_item, bytes):
                v_item = v_item.decode("utf-8")
            elif isinstance(v_item, dict):
                v_item = self._process_dict(v_item)
            elif isinstance(v_item, (list, tuple)):
                v_item = [
                    v1.decode("utf-8") if isinstance(v1, bytes) else v1 for v1 in v_item
                ]
            new_dict[k] = v_item
        return new_dict

    def print_result(self, data=None):
        """Print result object pretty.

        :param data: Data to be printed.
        """
        data = data or self.data
        if isinstance(data, dict):
            data = self._process_dict(data)
        elif isinstance(data, (list, tuple)):
            data = [v.decode("utf-8") if isinstance(v, bytes) else v for v in data]
        elif isinstance(data, bytes):
            data = data.decode("utf-8")
        tokens = list(pygments.lex(json.dumps(data, indent=4), lexer=JsonLexer()))
        print_formatted_text(PygmentsTokens(tokens))
